from hazelcast.protocol.builtin import FixSizedTypesCodec, CodecUtil
from hazelcast.serialization.bits import *
from hazelcast.protocol.client_message import END_FRAME_BUF, END_FINAL_FRAME_BUF, SIZE_OF_FRAME_LENGTH_AND_FLAGS, create_initial_buffer_custom
from hazelcast.core import Address
from hazelcast.protocol.builtin import StringCodec

_PORT_ENCODE_OFFSET = 2 * SIZE_OF_FRAME_LENGTH_AND_FLAGS
_PORT_DECODE_OFFSET = 0
_INITIAL_FRAME_SIZE = _PORT_ENCODE_OFFSET + INT_SIZE_IN_BYTES - SIZE_OF_FRAME_LENGTH_AND_FLAGS


class AddressCodec:
    @staticmethod
    def encode(buf, address, is_final=False):
        initial_frame_buf = create_initial_buffer_custom(_INITIAL_FRAME_SIZE)
        FixSizedTypesCodec.encode_int(initial_frame_buf, _PORT_ENCODE_OFFSET, address.port)
        buf.extend(initial_frame_buf)
        StringCodec.encode(buf, address.host)
        if is_final:
            buf.extend(END_FINAL_FRAME_BUF)
        else:
            buf.extend(END_FRAME_BUF)

    @staticmethod
    def decode(msg):
        msg.next_frame()
        initial_frame = msg.next_frame()
        port = FixSizedTypesCodec.decode_int(initial_frame.buf, _PORT_DECODE_OFFSET)
        host = StringCodec.decode(msg)
        CodecUtil.fast_forward_to_end_frame(msg)
        return Address(host, port)



import typing

from collections import defaultdict

from hazelcast.future import combine_futures, Future, ImmediateFuture
from hazelcast.protocol.codec import (
    multi_map_add_entry_listener_codec,
    multi_map_add_entry_listener_to_key_codec,
    multi_map_clear_codec,
    multi_map_contains_entry_codec,
    multi_map_contains_key_codec,
    multi_map_contains_value_codec,
    multi_map_entry_set_codec,
    multi_map_force_unlock_codec,
    multi_map_get_codec,
    multi_map_is_locked_codec,
    multi_map_key_set_codec,
    multi_map_lock_codec,
    multi_map_put_codec,
    multi_map_put_all_codec,
    multi_map_remove_codec,
    multi_map_remove_entry_codec,
    multi_map_remove_entry_listener_codec,
    multi_map_size_codec,
    multi_map_try_lock_codec,
    multi_map_unlock_codec,
    multi_map_value_count_codec,
    multi_map_values_codec,
)
from hazelcast.proxy.base import Proxy, EntryEvent, EntryEventType
from hazelcast.types import ValueType, KeyType
from hazelcast.serialization.data import Data
from hazelcast.serialization.compact import SchemaNotReplicatedError
from hazelcast.util import (
    check_not_none,
    thread_id,
    to_millis,
    deserialize_list_in_place,
    deserialize_entry_list_in_place,
)

EntryEventCallable = typing.Callable[[EntryEvent[KeyType, ValueType]], None]


class MultiMap(Proxy["BlockingMultiMap"], typing.Generic[KeyType, ValueType]):
    """A specialized map whose keys can be associated with multiple values."""

    def __init__(self, service_name, name, context):
        super(MultiMap, self).__init__(service_name, name, context)
        self._reference_id_generator = context.lock_reference_id_generator

    def add_entry_listener(
        self,
        include_value: bool = False,
        key: KeyType = None,
        added_func: EntryEventCallable = None,
        removed_func: EntryEventCallable = None,
        clear_all_func: EntryEventCallable = None,
    ) -> Future[str]:
        """Adds an entry listener for this multimap.

        The listener will be notified for all multimap add/remove/clear-all
        events.

        Args:
            include_value: Whether received event should include the value or
                not.
            key: Key for filtering the events.
            added_func: Function to be called when an entry is added to map.
            removed_func: Function to be called when an entry is removed from
                map.
            clear_all_func: Function to be called when entries are cleared
                from map.

        Returns:
            A registration id which is used as a key to remove the listener.
        """
        if key is not None:
            try:
                key_data = self._to_data(key)
            except SchemaNotReplicatedError as e:
                return self._send_schema_and_retry(
                    e,
                    self.add_entry_listener,
                    include_value,
                    key,
                    added_func,
                    removed_func,
                    clear_all_func,
                )

            with_key_codec = multi_map_add_entry_listener_to_key_codec
            request = with_key_codec.encode_request(self.name, key_data, include_value, False)
            response_decoder = with_key_codec.decode_response
            event_message_handler = with_key_codec.handle
        else:
            codec = multi_map_add_entry_listener_codec
            request = codec.encode_request(self.name, include_value, False)
            response_decoder = codec.decode_response
            event_message_handler = codec.handle

        def handle_event_entry(
            key_data,
            value_data,
            old_value_data,
            merging_value_data,
            event_type,
            uuid,
            number_of_affected_entries,
        ):
            event = EntryEvent(
                self._to_object(key_data),
                self._to_object(value_data),
                self._to_object(old_value_data),
                self._to_object(merging_value_data),
                event_type,
                uuid,
                number_of_affected_entries,
            )
            if event.event_type == EntryEventType.ADDED and added_func:
                added_func(event)
            elif event.event_type == EntryEventType.REMOVED and removed_func:
                removed_func(event)
            elif event.event_type == EntryEventType.CLEAR_ALL and clear_all_func:
                clear_all_func(event)

        return self._register_listener(
            request,
            lambda r: response_decoder(r),
            lambda reg_id: multi_map_remove_entry_listener_codec.encode_request(self.name, reg_id),
            lambda m: event_message_handler(m, handle_event_entry),
        )

    def contains_key(self, key: KeyType) -> Future[bool]:
        """Determines whether this multimap contains an entry with the key.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form
            of the key, not the actual implementations of ``__hash__`` and
            ``__eq__`` defined in key's class.

        Args:
            key: The specified key.

        Returns:
            ``True`` if this multimap contains an entry for the specified key,
            ``False`` otherwise.
        """
        check_not_none(key, "key can't be None")
        try:
            key_data = self._to_data(key)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.contains_key, key)

        request = multi_map_contains_key_codec.encode_request(self.name, key_data, thread_id())
        return self._invoke_on_key(request, key_data, multi_map_contains_key_codec.decode_response)

    def contains_value(self, value: ValueType) -> Future[bool]:
        """Determines whether this map contains one or more keys for the
        specified value.

        Args:
            value: The specified value.

        Returns:
            ``True`` if this multimap contains an entry for the specified
            value, ``False`` otherwise.
        """
        check_not_none(value, "value can't be None")
        try:
            value_data = self._to_data(value)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.contains_value, value)

        request = multi_map_contains_value_codec.encode_request(self.name, value_data)
        return self._invoke(request, multi_map_contains_value_codec.decode_response)

    def contains_entry(self, key: KeyType, value: ValueType) -> Future[bool]:
        """Returns whether the multimap contains an entry with the value.

        Args:
            key: The specified key.
            value: The specified value.

        Returns:
            ``True`` if this multimap contains the key-value tuple, ``False``
            otherwise.
        """
        check_not_none(key, "key can't be None")
        check_not_none(value, "value can't be None")
        try:
            key_data = self._to_data(key)
            value_data = self._to_data(value)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.contains_entry, key, value)

        request = multi_map_contains_entry_codec.encode_request(
            self.name, key_data, value_data, thread_id()
        )
        return self._invoke_on_key(
            request, key_data, multi_map_contains_entry_codec.decode_response
        )

    def clear(self) -> Future[None]:
        """Clears the multimap. Removes all key-value tuples."""
        request = multi_map_clear_codec.encode_request(self.name)
        return self._invoke(request)

    def entry_set(self) -> Future[typing.List[typing.Tuple[KeyType, ValueType]]]:
        """Returns the list of key-value tuples in the multimap.

        Warning:
            The list is NOT backed by the map, so changes to the map are NOT
            reflected in the list, and vice-versa.

        Returns:
            The list of key-value tuples in the multimap.
        """

        def handler(message):
            entry_data_list = multi_map_entry_set_codec.decode_response(message)
            return deserialize_entry_list_in_place(entry_data_list, self._to_object)

        request = multi_map_entry_set_codec.encode_request(self.name)
        return self._invoke(request, handler)

    def get(self, key: KeyType) -> Future[typing.Optional[typing.List[ValueType]]]:
        """Returns the list of values associated with the key. ``None`` if
        this map does not contain this key.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` of the binary form of
            the key, not the actual implementations of ``__hash__`` and
            ``__eq__`` defined in the key's class.

        Warning:
            The list is NOT backed by the multimap, so changes to the map are
            list reflected in the collection, and vice-versa.

        Args:
            key: The specified key.

        Returns:
            The list of the values associated with the specified key.
        """
        check_not_none(key, "key can't be None")
        try:
            key_data = self._to_data(key)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.get, key)

        def handler(message):
            data_list = multi_map_get_codec.decode_response(message)
            return deserialize_list_in_place(data_list, self._to_object)

        request = multi_map_get_codec.encode_request(self.name, key_data, thread_id())
        return self._invoke_on_key(request, key_data, handler)

    def is_locked(self, key: KeyType) -> Future[bool]:
        """Checks the lock for the specified key.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form
            of the key, not the actual implementations of ``__hash__`` and
            ``__eq__`` defined in key's class.

        Args:
            key: The key that is checked for lock.

        Returns:
            ``True`` if lock is acquired, ``False`` otherwise.
        """
        check_not_none(key, "key can't be None")
        try:
            key_data = self._to_data(key)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.is_locked, key)

        request = multi_map_is_locked_codec.encode_request(self.name, key_data)
        return self._invoke_on_key(request, key_data, multi_map_is_locked_codec.decode_response)

    def force_unlock(self, key: KeyType) -> Future[None]:
        """Releases the lock for the specified key regardless of the lock
        owner.

        It always successfully unlocks the key, never blocks, and returns
        immediately.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form
            of the key, not the actual implementations of ``__hash__`` and
            ``__eq__`` defined in key's class.

        Args:
            key: The key to lock.
        """
        check_not_none(key, "key can't be None")
        try:
            key_data = self._to_data(key)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.force_unlock, key)

        request = multi_map_force_unlock_codec.encode_request(
            self.name, key_data, self._reference_id_generator.get_and_increment()
        )
        return self._invoke_on_key(request, key_data)

    def key_set(self) -> Future[typing.List[KeyType]]:
        """Returns the list of keys in the multimap.

        Warning:
            The list is NOT backed by the map, so changes to the map are NOT
            reflected in the list, and vice-versa.

        Returns:
            A list of the clone of the keys.
        """

        def handler(message):
            data_list = multi_map_key_set_codec.decode_response(message)
            return deserialize_list_in_place(data_list, self._to_object)

        request = multi_map_key_set_codec.encode_request(self.name)
        return self._invoke(request, handler)

    def lock(self, key: KeyType, lease_time: float = None) -> Future[None]:
        """Acquires the lock for the specified key infinitely or for the
        specified lease time if provided.

        If the lock is not available, the current thread becomes disabled for
        thread scheduling purposes and lies dormant until the lock has been
        acquired.

        Scope of the lock is this map only. Acquired lock is only for the key
        in this map.

        Locks are re-entrant; so, if the key is locked N times, it should be
        unlocked N times before another thread can acquire it.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form
            of the key, not the actual implementations of ``__hash__`` and
            ``__eq__`` defined in key's class.

        Args:
            key: The key to lock.
            lease_time: Time in seconds to wait before releasing the lock.
        """
        check_not_none(key, "key can't be None")
        try:
            key_data = self._to_data(key)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.lock, key, lease_time)

        request = multi_map_lock_codec.encode_request(
            self.name,
            key_data,
            thread_id(),
            to_millis(lease_time),
            self._reference_id_generator.get_and_increment(),
        )
        return self._invoke_on_key(request, key_data)

    def remove(self, key: KeyType, value: ValueType) -> Future[bool]:
        """Removes the given key-value tuple from the multimap.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form
            of the key, not the actual implementations of ``__hash__`` and
            ``__eq__`` defined in key's class.

        Args:
            key: The key of the entry to remove.
            value: The value of the entry to remove.

        Returns:
            ``True`` if the size of the multimap changed after the remove
            operation, ``False`` otherwise.
        """
        check_not_none(key, "key can't be None")
        check_not_none(key, "value can't be None")
        try:
            key_data = self._to_data(key)
            value_data = self._to_data(value)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.remove, key, value)

        request = multi_map_remove_entry_codec.encode_request(
            self.name, key_data, value_data, thread_id()
        )
        return self._invoke_on_key(request, key_data, multi_map_remove_entry_codec.decode_response)

    def remove_all(self, key: KeyType) -> Future[typing.List[ValueType]]:
        """Removes all the entries with the given key and returns the value
        list associated with this key.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form
            of the key, not the actual implementations of ``__hash__`` and
            ``__eq__`` defined in key's class.

        Warning:
            The returned list is NOT backed by the map, so changes to the map
            are NOT reflected in the list, and vice-versa.

        Args:
            key: The key of the entries to remove.

        Returns:
            The collection of removed values associated with the given key.
        """
        check_not_none(key, "key can't be None")

        def handler(message):
            data_list = multi_map_remove_codec.decode_response(message)
            return deserialize_list_in_place(data_list, self._to_object)

        try:
            key_data = self._to_data(key)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.remove_all, key)

        request = multi_map_remove_codec.encode_request(self.name, key_data, thread_id())
        return self._invoke_on_key(request, key_data, handler)

    def put(self, key: KeyType, value: ValueType) -> Future[bool]:
        """Stores a key-value tuple in the multimap.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form
            of the key, not the actual implementations of ``__hash__`` and
            ``__eq__`` defined in key's class.

        Args:
            key: The key to be stored.
            value: The value to be stored.

        Returns:
            ``True`` if size of the multimap is increased, ``False`` if the
            multimap already contains the key-value tuple.
        """
        check_not_none(key, "key can't be None")
        check_not_none(value, "value can't be None")
        try:
            key_data = self._to_data(key)
            value_data = self._to_data(value)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.put, key, value)

        request = multi_map_put_codec.encode_request(self.name, key_data, value_data, thread_id())
        return self._invoke_on_key(request, key_data, multi_map_put_codec.decode_response)

    def put_all(self, multimap: typing.Dict[KeyType, typing.Sequence[ValueType]]) -> Future[None]:
        """Stores the given Map in the MultiMap.

        The results of concurrently mutating the given map are undefined.
        No atomicity guarantees are given. It could be that in case of failure some of the key/value-pairs get written, while others are not.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form
            of the key, not the actual implementations of ``__hash__`` and
            ``__eq__`` defined in key's class.

        Args:
            multimap: the map corresponds to multimap entries.
        """
        check_not_none(multimap, "multimap can't be None")
        if not multimap:
            return ImmediateFuture(None)

        partition_service = self._context.partition_service
        partition_map: typing.DefaultDict[
            int, typing.List[typing.Tuple[Data, typing.List[Data]]]
        ] = defaultdict(list)

        for key, values in multimap.items():
            try:
                check_not_none(key, "key can't be None")
                check_not_none(values, "values can't be None")
                serialized_key = self._to_data(key)
                serialized_values = []
                for value in values:
                    check_not_none(value, "value can't be None")
                    serialized_values.append(self._to_data(value))
                partition_id = partition_service.get_partition_id(serialized_key)
                partition_map[partition_id].append((serialized_key, serialized_values))
            except SchemaNotReplicatedError as e:
                return self._send_schema_and_retry(e, self.put_all, multimap)

        futures = []
        for partition_id, entry_list in partition_map.items():
            request = multi_map_put_all_codec.encode_request(self.name, entry_list)
            future = self._invoke_on_partition(request, partition_id)
            futures.append(future)
        return combine_futures(futures).continue_with(lambda _: None)

    def remove_entry_listener(self, registration_id: str) -> Future[bool]:
        """Removes the specified entry listener.

        Returns silently if there is no such listener added before.

        Args:
            registration_id: Id of registered listener.

        Returns:
            ``True`` if registration is removed, ``False`` otherwise.
        """
        return self._deregister_listener(registration_id)

    def size(self) -> Future[int]:
        """Returns the number of entries in this multimap.

        Returns:
            Number of entries in this multimap.
        """
        request = multi_map_size_codec.encode_request(self.name)
        return self._invoke(request, multi_map_size_codec.decode_response)

    def value_count(self, key: KeyType) -> Future[int]:
        """Returns the number of values that match the given key in the
        multimap.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form
            of the key, not the actual implementations of ``__hash__`` and
            ``__eq__`` defined in key's class.

        Args:
            key: The key whose values count is to be returned.

        Returns:
            The number of values that match the given key in the multimap.
        """
        check_not_none(key, "key can't be None")
        try:
            key_data = self._to_data(key)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.value_count, key)

        request = multi_map_value_count_codec.encode_request(self.name, key_data, thread_id())
        return self._invoke_on_key(request, key_data, multi_map_value_count_codec.decode_response)

    def values(self) -> Future[typing.List[ValueType]]:
        """Returns the list of values in the multimap.

        Warning:
            The returned list is NOT backed by the map, so changes to the map
            are NOT reflected in the list, and vice-versa.

        Returns:
            The list of values in the multimap.
        """

        def handler(message):
            data_list = multi_map_values_codec.decode_response(message)
            return deserialize_list_in_place(data_list, self._to_object)

        request = multi_map_values_codec.encode_request(self.name)
        return self._invoke(request, handler)

    def try_lock(self, key: KeyType, lease_time: float = None, timeout: float = 0) -> Future[bool]:
        """Tries to acquire the lock for the specified key.

        When the lock is not available:

        - If the timeout is not provided, the current thread doesn't wait and
          returns ``False`` immediately.
        - If the timeout is provided, the current thread becomes disabled for
          thread scheduling purposes and lies dormant until one of the
          followings happens:

            - The lock is acquired by the current thread, or
            - The specified waiting time elapses.

        If the lease time is provided, lock will be released after this time
        elapses.

        Args:
            key: Key to lock in this map.
            lease_time: Time in seconds to wait before releasing the lock.
            timeout: Maximum time in seconds to wait for the lock.

        Returns:
            ``True`` if the lock was acquired, ``False`` otherwise.
        """
        check_not_none(key, "key can't be None")
        try:
            key_data = self._to_data(key)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.try_lock, key, lease_time, timeout)

        request = multi_map_try_lock_codec.encode_request(
            self.name,
            key_data,
            thread_id(),
            to_millis(lease_time),
            to_millis(timeout),
            self._reference_id_generator.get_and_increment(),
        )
        return self._invoke_on_key(request, key_data, multi_map_try_lock_codec.decode_response)

    def unlock(self, key: KeyType) -> Future[None]:
        """Releases the lock for the specified key. It never blocks and
        returns immediately.

        Warning:
            This method uses ``__hash__`` and ``__eq__`` methods of binary form
            of the key, not the actual implementations of ``__hash__`` and
            ``__eq__`` defined in key's class.

        Args:
            key: The key to lock.
        """
        check_not_none(key, "key can't be None")
        try:
            key_data = self._to_data(key)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.unlock, key)

        request = multi_map_unlock_codec.encode_request(
            self.name, key_data, thread_id(), self._reference_id_generator.get_and_increment()
        )
        return self._invoke_on_key(request, key_data)

    def blocking(self) -> "BlockingMultiMap[KeyType, ValueType]":
        return BlockingMultiMap(self)


class BlockingMultiMap(MultiMap[KeyType, ValueType]):
    __slots__ = ("_wrapped", "name", "service_name")

    def __init__(self, wrapped: MultiMap[KeyType, ValueType]):
        self.name = wrapped.name
        self.service_name = wrapped.service_name
        self._wrapped = wrapped

    def add_entry_listener(  # type: ignore[override]
        self,
        include_value: bool = False,
        key: KeyType = None,
        added_func: EntryEventCallable = None,
        removed_func: EntryEventCallable = None,
        clear_all_func: EntryEventCallable = None,
    ) -> str:
        return self._wrapped.add_entry_listener(
            include_value, key, added_func, removed_func, clear_all_func
        ).result()

    def contains_key(  # type: ignore[override]
        self,
        key: KeyType,
    ) -> bool:
        return self._wrapped.contains_key(key).result()

    def contains_value(  # type: ignore[override]
        self,
        value: ValueType,
    ) -> bool:
        return self._wrapped.contains_value(value).result()

    def contains_entry(  # type: ignore[override]
        self,
        key: KeyType,
        value: ValueType,
    ) -> bool:
        return self._wrapped.contains_entry(key, value).result()

    def clear(  # type: ignore[override]
        self,
    ) -> None:
        return self._wrapped.clear().result()

    def entry_set(  # type: ignore[override]
        self,
    ) -> typing.List[typing.Tuple[KeyType, ValueType]]:
        return self._wrapped.entry_set().result()

    def get(  # type: ignore[override]
        self,
        key: KeyType,
    ) -> typing.Optional[typing.List[ValueType]]:
        return self._wrapped.get(key).result()

    def is_locked(  # type: ignore[override]
        self,
        key: KeyType,
    ) -> bool:
        return self._wrapped.is_locked(key).result()

    def force_unlock(  # type: ignore[override]
        self,
        key: KeyType,
    ) -> None:
        return self._wrapped.force_unlock(key).result()

    def key_set(  # type: ignore[override]
        self,
    ) -> typing.List[KeyType]:
        return self._wrapped.key_set().result()

    def lock(  # type: ignore[override]
        self,
        key: KeyType,
        lease_time: float = None,
    ) -> None:
        return self._wrapped.lock(key, lease_time).result()

    def remove(  # type: ignore[override]
        self,
        key: KeyType,
        value: ValueType,
    ) -> bool:
        return self._wrapped.remove(key, value).result()

    def remove_all(  # type: ignore[override]
        self,
        key: KeyType,
    ) -> typing.List[ValueType]:
        return self._wrapped.remove_all(key).result()

    def put(  # type: ignore[override]
        self,
        key: KeyType,
        value: ValueType,
    ) -> bool:
        return self._wrapped.put(key, value).result()

    def put_all(  # type: ignore[override]
        self, multimap: typing.Dict[KeyType, typing.Sequence[ValueType]]
    ) -> None:
        self._wrapped.put_all(multimap).result()

    def remove_entry_listener(  # type: ignore[override]
        self,
        registration_id: str,
    ) -> bool:
        return self._wrapped.remove_entry_listener(registration_id).result()

    def size(  # type: ignore[override]
        self,
    ) -> int:
        return self._wrapped.size().result()

    def value_count(  # type: ignore[override]
        self,
        key: KeyType,
    ) -> int:
        return self._wrapped.value_count(key).result()

    def values(  # type: ignore[override]
        self,
    ) -> typing.List[ValueType]:
        return self._wrapped.values().result()

    def try_lock(  # type: ignore[override]
        self,
        key: KeyType,
        lease_time: float = None,
        timeout: float = 0,
    ) -> bool:
        return self._wrapped.try_lock(key, lease_time, timeout).result()

    def unlock(  # type: ignore[override]
        self,
        key: KeyType,
    ) -> None:
        return self._wrapped.unlock(key).result()

    def destroy(self) -> bool:
        return self._wrapped.destroy()

    def blocking(self) -> "BlockingMultiMap[KeyType, ValueType]":
        return self

    def __repr__(self) -> str:
        return self._wrapped.__repr__()



"""Hazelcast Core objects and constants."""
import json
import typing
import uuid

from hazelcast.types import KeyType, ValueType

CLIENT_TYPE = "PYH"
SERIALIZATION_VERSION = 1


class MemberInfo:
    """
    Represents a member in the cluster with its address, uuid, lite member
    status, attributes, version, and address map.
    """

    __slots__ = ("address", "uuid", "attributes", "lite_member", "version", "address_map")

    def __init__(
        self,
        address: "Address",
        member_uuid: uuid.UUID,
        attributes: typing.Dict[str, str],
        lite_member: bool,
        version: "MemberVersion",
        _,
        address_map: typing.Dict["EndpointQualifier", "Address"],
    ):
        self.address = address
        """
        Address of the member.
        """

        self.uuid = member_uuid
        """
        UUID of the member.
        """

        self.attributes = attributes
        """
        Configured attributes of the member.
        """

        self.lite_member = lite_member
        """
        ``True`` if the member is a lite member, ``False`` otherwise. Lite 
        members do not own any partition.
        """

        self.version = version
        """
        Hazelcast codebase version of the member.
        """

        self.address_map = address_map
        """
        Dictionary of server socket addresses per :class:`EndpointQualifier` 
        of this member.
        """

    def __str__(self):
        return "Member [%s]:%s - %s%s" % (
            self.address.host,
            self.address.port,
            self.uuid,
            " lite" if self.lite_member else "",
        )

    def __repr__(self):
        return "Member(address=%s, uuid=%s, attributes=%s, lite_member=%s, version=%s)" % (
            self.address,
            self.uuid,
            self.attributes,
            self.lite_member,
            self.version,
        )

    def __hash__(self):
        return hash((self.address, self.uuid))

    def __eq__(self, other):
        return (
            isinstance(other, MemberInfo)
            and self.address == other.address
            and self.uuid == other.uuid
        )

    def __ne__(self, other):
        return not self.__eq__(other)


class Address:
    """Represents an address of a member in the cluster."""

    def __init__(self, host: str, port: int):
        self.host = host
        """
        Host of the address.
        """

        self.port = port
        """
        Port of the address.
        """

    def __repr__(self):
        return "Address(host=%s, port=%d)" % (self.host, self.port)

    def __hash__(self):
        return hash((self.host, self.port))

    def __eq__(self, other):
        return isinstance(other, Address) and self.host == other.host and self.port == other.port

    def __ne__(self, other):
        return not self.__eq__(other)


class ProtocolType:
    """Types of server sockets.

    A member typically responds to several types of protocols for
    member-to-member, client-member protocol, WAN communication etc. The
    default configuration uses a single server socket to listen for all kinds
    of protocol types configured, while Advanced Network Config of the server
    allows configuration of multiple server sockets.
    """

    # We had to put dummy documentations for the constants
    # so that they are displayed on the API documentation.

    MEMBER = 0
    """Type of member server sockets."""

    CLIENT = 1
    """Type of client server sockets."""

    WAN = 2
    """Type of WAN server sockets."""

    REST = 3
    """Type of REST server sockets."""

    MEMCACHE = 4
    """Type of Memcached server sockets."""


class EndpointQualifier:
    """Uniquely identifies groups of network connections sharing a common
    :class:`ProtocolType` and the same network settings, when Hazelcast server
    is configured with Advanced Network Configuration enabled.

    In some cases, just the :class:`ProtocolType` is enough (e.g. since there
    can be only a single member server socket).

    When just the :class:`ProtocolType` is not enough (for example when
    configuring outgoing WAN connections to 2 different target clusters),
    an :attr:`identifier` is used to uniquely identify the network
    configuration.
    """

    __slots__ = ("_protocol_type", "_identifier")

    def __init__(self, protocol_type: int, identifier: typing.Optional[str]):
        self._protocol_type = protocol_type
        self._identifier = identifier

    @property
    def protocol_type(self) -> int:
        """Protocol type of the endpoint."""
        return self._protocol_type

    @property
    def identifier(self) -> typing.Optional[str]:
        """Unique identifier for same-protocol-type endpoints."""
        return self._identifier

    def __eq__(self, other):
        return (
            isinstance(other, EndpointQualifier)
            and self._protocol_type == other._protocol_type
            and self._identifier == other._identifier
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash((self._protocol_type, self._identifier))

    def __repr__(self):
        return "EndpointQualifier(protocol_type=%s, identifier=%s)" % (
            self._protocol_type,
            self._identifier,
        )


class AddressHelper:
    @staticmethod
    def get_possible_addresses(address):
        address = AddressHelper.address_from_str(address)
        possible_port = address.port
        port_try_count = 1
        if possible_port == -1:
            port_try_count = 3
            possible_port = 5701

        addresses = []
        for i in range(port_try_count):
            addresses.append(Address(address.host, possible_port + i))

        # primary, secondary
        return [addresses.pop(0)], addresses

    @staticmethod
    def address_from_str(address, port=-1):
        bracket_start_idx = address.find("[")
        bracket_end_idx = address.find("]", bracket_start_idx)
        colon_idx = address.find(":")
        last_colon_idx = address.rfind(":")

        if -1 < colon_idx < last_colon_idx:
            # IPv6
            if bracket_start_idx == 0 and bracket_end_idx > bracket_start_idx:
                host = address[bracket_start_idx + 1 : bracket_end_idx]
                if last_colon_idx == (bracket_end_idx + 1):
                    port = int(address[last_colon_idx + 1 :])
            else:
                host = address
        elif colon_idx > 0 and colon_idx == last_colon_idx:
            host = address[:colon_idx]
            port = int(address[colon_idx + 1 :])
        else:
            host = address
        return Address(host, port)


class DistributedObjectInfo:
    def __init__(self, service_name, name):
        self.service_name = service_name
        self.name = name

    def __repr__(self):
        return "DistributedObjectInfo(serviceName=%s, name=%s)" % (self.service_name, self.name)

    def __hash__(self):
        return hash((self.name, self.service_name))

    def __eq__(self, other):
        if isinstance(other, DistributedObjectInfo):
            return self.name == other.name and self.service_name == other.service_name
        return False


class DistributedObjectEventType:
    """Type of the distributed object event."""

    CREATED = "CREATED"
    """
    DistributedObject is created.
    """

    DESTROYED = "DESTROYED"
    """
    DistributedObject is destroyed.
    """


class DistributedObjectEvent:
    """Distributed Object Event"""

    def __init__(self, name: str, service_name: str, event_type: str, source: uuid.UUID):
        self.name = name
        """
        Name of the distributed object.
        """

        self.service_name = service_name
        """
        Service name of the distributed object.
        """

        self.event_type = event_type
        """
        Event type. Either ``CREATED`` or ``DESTROYED``.
        """

        self.source = source
        """
        UUID of the member that fired the event.
        """

    def __repr__(self):
        return "DistributedObjectEvent(name=%s, service_name=%s, event_type=%s, source=%s)" % (
            self.name,
            self.service_name,
            self.event_type,
            self.source,
        )


class SimpleEntryView(typing.Generic[KeyType, ValueType]):
    """EntryView represents a readonly view of a map entry."""

    def __init__(
        self,
        key: KeyType,
        value: ValueType,
        cost: int,
        creation_time: int,
        expiration_time: int,
        hits: int,
        last_access_time: int,
        last_stored_time: int,
        last_update_time: int,
        version: int,
        ttl: int,
        max_idle: int,
    ):
        self.key = key
        """
        The key of the entry.
        """

        self.value = value
        """
        The value of the entry.
        """

        self.cost = cost
        """
        The cost in bytes of the entry.
        """

        self.creation_time = creation_time
        """
        The creation time of the entry.
        """

        self.expiration_time = expiration_time
        """
        The expiration time of the entry.
        """

        self.hits = hits
        """
        Number of hits of the entry.
        """

        self.last_access_time = last_access_time
        """
        The last access time for the entry.
        """

        self.last_stored_time = last_stored_time
        """
        The last store time for the value.
        """

        self.last_update_time = last_update_time
        """
        The last time the value was updated.
        """

        self.version = version
        """
        The version of the entry.
        """

        self.ttl = ttl
        """
        The last set time to live milliseconds.
        """

        self.max_idle = max_idle
        """
        The last set max idle time in milliseconds.
        """

    def __repr__(self):
        return (
            "SimpleEntryView(key=%s, value=%s, cost=%s, creation_time=%s, "
            "expiration_time=%s, hits=%s, last_access_time=%s, last_stored_time=%s, "
            "last_update_time=%s, version=%s, ttl=%s, max_idle=%s"
            % (
                self.key,
                self.value,
                self.cost,
                self.creation_time,
                self.expiration_time,
                self.hits,
                self.last_access_time,
                self.last_stored_time,
                self.last_update_time,
                self.version,
                self.ttl,
                self.max_idle,
            )
        )


class HazelcastJsonValue:
    """HazelcastJsonValue is a wrapper for JSON formatted strings.

    It is preferred to store HazelcastJsonValue instead of Strings for JSON
    formatted strings. Users can run predicates and use indexes on the
    attributes of the underlying JSON strings.

    HazelcastJsonValue is queried using Hazelcast's querying language.

    In terms of querying, numbers in JSON strings are treated as either
    Long or Double in the Java side. str, bool and None
    are treated as String, boolean and null respectively.

    HazelcastJsonValue keeps given string as it is. Strings are not
    checked for being valid. Ill-formatted JSON strings may cause false
    positive or false negative results in queries.

    HazelcastJsonValue can also be constructed from JSON serializable objects.
    In that case, objects are converted to JSON strings and stored as such.
    If an error occurs during the conversion, it is raised directly.

    None values are not allowed.
    """

    def __init__(self, value: typing.Any):
        if value is None:
            raise AssertionError("JSON string or the object cannot be None.")
        if isinstance(value, str):
            self._json_string = value
        else:
            self._json_string = json.dumps(value)

    def to_string(self) -> str:
        """Returns unaltered string that was used to create this object.

        Returns:
            The original string.
        """
        return self._json_string

    def loads(self) -> typing.Any:
        """Deserializes the string that was used to create this object
        and returns as Python object.

        Returns:
            The Python object represented by the original string.
        """
        return json.loads(self._json_string)

    def __eq__(self, other):
        return isinstance(other, HazelcastJsonValue) and self._json_string == other._json_string

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self._json_string)

    def __repr__(self):
        return self._json_string


class MemberVersion:
    """
    Determines the Hazelcast codebase version in terms of major.minor.patch version.
    """

    __slots__ = ("major", "minor", "patch")

    def __init__(self, major: int, minor: int, patch: int):
        self.major = major
        self.minor = minor
        self.patch = patch

    def __repr__(self):
        return "MemberVersion(major=%s, minor=%s, patch=%s)" % (self.major, self.minor, self.patch)


class MapEntry(typing.Generic[KeyType, ValueType]):
    """
    Represents the entry of a Map, with key and value fields.
    """

    __slots__ = ("_key", "_value")

    def __init__(self, key: KeyType = None, value: ValueType = None):
        self._key = key
        self._value = value

    @property
    def key(self) -> KeyType:
        """Key of the entry."""
        # This has the correct type, but due to structure of the identified
        # data serialization, the constructor must have ``None`` default
        # values, which upsets the type checker.
        return self._key  # type: ignore[return-value]

    @property
    def value(self) -> ValueType:
        """Value of the entry."""
        # This has the correct type, but due to structure of the identified
        # data serialization, the constructor must have ``None`` default
        # values, which upsets the type checker.
        return self._value  # type: ignore[return-value]



__version__ = "5.2.0"

# Set the default handler to "hazelcast" loggers
# to avoid "No handlers could be found" warnings.
import logging

logging.getLogger(__name__).addHandler(logging.NullHandler())

from hazelcast.client import HazelcastClient



from hazelcast.serialization.api import *
from hazelcast.serialization.bits import *


class _ObjectDataInput(ObjectDataInput):
    __slots__ = (
        "_buffer",
        "_service",
        "_is_big_endian",
        "_pos",
        "_size",
        "_FMT_INT8",
        "_FMT_INT",
        "_FMT_SHORT",
        "_FMT_CHAR",
        "_FMT_LONG",
        "_FMT_FLOAT",
        "_FMT_DOUBLE",
    )

    def __init__(self, buff, offset=0, serialization_service=None, is_big_endian=True):
        self._buffer = buff
        self._service = serialization_service
        self._is_big_endian = is_big_endian
        self._pos = offset
        self._size = len(buff)
        # Local cache struct formats according to endianness
        self._FMT_INT8 = BE_INT8 if self._is_big_endian else LE_INT8
        self._FMT_INT = BE_INT if self._is_big_endian else LE_INT
        self._FMT_SHORT = BE_INT16 if self._is_big_endian else LE_INT16
        self._FMT_CHAR = BE_UINT16 if self._is_big_endian else LE_UINT16
        self._FMT_LONG = BE_LONG if self._is_big_endian else LE_LONG
        self._FMT_FLOAT = BE_FLOAT if self._is_big_endian else LE_FLOAT
        self._FMT_DOUBLE = BE_DOUBLE if self._is_big_endian else LE_DOUBLE

    def read_into(self, buff, offset=None, length=None):
        _off = offset if offset is not None else 0
        _len = length if length is not None else len(buff)
        if _off < 0 or _len < 0 or (_off + _len) > len(self._buffer):
            raise IndexError()
        elif _len == 0:
            return
        if self._pos > self._size:
            raise IndexError()
        if self._pos + _len > self._size:
            _len = self._size - self._pos
        buff[_off : _off + _len] = self._buffer[self._pos : self._pos + _len]
        self._pos += _len

    def skip_bytes(self, count):
        if count <= 0:
            return 0

        if self._pos + count > self._size:
            count = self._size - self._pos

        self._pos += count
        return count

    def read_boolean(self):
        return self.read_byte() != 0

    def read_boolean_positional(self, position: int) -> bool:
        return self.read_byte_positional(position) != 0

    def read_byte(self):
        self._check_available(self._pos, BYTE_SIZE_IN_BYTES)
        value = self._FMT_INT8.unpack_from(self._buffer, self._pos)[0]
        self._pos += BYTE_SIZE_IN_BYTES
        return value

    def read_byte_positional(self, position: int) -> int:
        self._check_available(position, BYTE_SIZE_IN_BYTES)
        return self._FMT_INT8.unpack_from(self._buffer, position)[0]

    def read_unsigned_byte(self):
        self._check_available(self._pos, BYTE_SIZE_IN_BYTES)
        value = self._buffer[self._pos]
        self._pos += BYTE_SIZE_IN_BYTES
        return value

    def read_char(self):
        char_ord = self.read_short()
        return chr(char_ord)

    def read_char_positional(self, position: int) -> str:
        char_ord = self.read_short_positional(position)
        return chr(char_ord)

    def read_short(self):
        self._check_available(self._pos, SHORT_SIZE_IN_BYTES)
        value = self._FMT_SHORT.unpack_from(self._buffer, self._pos)[0]
        self._pos += SHORT_SIZE_IN_BYTES
        return value

    def read_short_positional(self, position: int) -> int:
        self._check_available(position, SHORT_SIZE_IN_BYTES)
        return self._FMT_SHORT.unpack_from(self._buffer, position)[0]

    def read_unsigned_short(self):
        self._check_available(self._pos, SHORT_SIZE_IN_BYTES)
        value = self._FMT_CHAR.unpack_from(self._buffer, self._pos)[0]
        self._pos += SHORT_SIZE_IN_BYTES
        return value

    def read_int(self):
        self._check_available(self._pos, INT_SIZE_IN_BYTES)
        value = self._FMT_INT.unpack_from(self._buffer, self._pos)[0]
        self._pos += INT_SIZE_IN_BYTES
        return value

    def read_int_positional(self, position: int) -> int:
        self._check_available(position, INT_SIZE_IN_BYTES)
        return self._FMT_INT.unpack_from(self._buffer, position)[0]

    def read_long(self):
        self._check_available(self._pos, LONG_SIZE_IN_BYTES)
        value = self._FMT_LONG.unpack_from(self._buffer, self._pos)[0]
        self._pos += LONG_SIZE_IN_BYTES
        return value

    def read_long_positional(self, position: int) -> int:
        self._check_available(position, LONG_SIZE_IN_BYTES)
        return self._FMT_LONG.unpack_from(self._buffer, position)[0]

    def read_float(self):
        self._check_available(self._pos, FLOAT_SIZE_IN_BYTES)
        value = self._FMT_FLOAT.unpack_from(self._buffer, self._pos)[0]
        self._pos += FLOAT_SIZE_IN_BYTES
        return value

    def read_float_positional(self, position: int) -> float:
        self._check_available(position, FLOAT_SIZE_IN_BYTES)
        return self._FMT_FLOAT.unpack_from(self._buffer, position)[0]

    def read_double(self):
        self._check_available(self._pos, DOUBLE_SIZE_IN_BYTES)
        value = self._FMT_DOUBLE.unpack_from(self._buffer, self._pos)[0]
        self._pos += DOUBLE_SIZE_IN_BYTES
        return value

    def read_double_positional(self, position: int) -> float:
        self._check_available(position, DOUBLE_SIZE_IN_BYTES)
        return self._FMT_DOUBLE.unpack_from(self._buffer, position)[0]

    def read_string(self):
        length = self.read_int()
        if length == NULL_ARRAY_LENGTH:
            return None
        result = bytearray(length)
        if length > 0:
            self.read_into(result, 0, length)
        return result.decode("utf-8")

    def read_byte_array(self):
        length = self.read_int()
        if length == NULL_ARRAY_LENGTH:
            return None
        result = bytearray(length)
        if length > 0:
            self.read_into(result, 0, length)

        return result

    def read_i8_array(self) -> typing.List[int]:
        return self._read_array_fnc(self.read_byte)

    def read_boolean_array(self):
        return self._read_array_fnc(self.read_boolean)

    def read_char_array(self):
        return self._read_array_fnc(self.read_char)

    def read_int_array(self):
        return self._read_array_fnc(self.read_int)

    def read_long_array(self):
        return self._read_array_fnc(self.read_long)

    def read_double_array(self):
        return self._read_array_fnc(self.read_double)

    def read_float_array(self):
        return self._read_array_fnc(self.read_float)

    def read_short_array(self):
        return self._read_array_fnc(self.read_short)

    def read_string_array(self):
        return self._read_array_fnc(self.read_string)

    def read_object(self):
        return self._service.read_object(self)

    def is_big_endian(self):
        return self._is_big_endian

    def position(self):
        return self._pos

    def set_position(self, position):
        self._pos = position

    def size(self):
        return self._size

    def get_byte_order(self):
        if self._is_big_endian:
            return "BIG_ENDIAN"
        return "LITTLE_ENDIAN"

    def read_utf(self):
        return self.read_string()

    def read_utf_array(self):
        return self.read_string_array()

    # HELPERS
    def _check_available(self, position, size):
        if position < 0:
            raise ValueError
        if self._size - position < size:
            raise EOFError("Cannot read %s bytes!" % size)

    def _read_array_fnc(self, read_item_fnc):
        length = self.read_int()
        if length == NULL_ARRAY_LENGTH:
            return None
        return [read_item_fnc() for _ in range(length)]



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer
from hazelcast.protocol.builtin import StringCodec
from hazelcast.protocol.builtin import ListMultiFrameCodec
from hazelcast.protocol.builtin import DataCodec

# hex: 0x012100
_REQUEST_MESSAGE_TYPE = 73984
# hex: 0x012101
_RESPONSE_MESSAGE_TYPE = 73985

_REQUEST_REPLACE_EXISTING_VALUES_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_REPLACE_EXISTING_VALUES_OFFSET + BOOLEAN_SIZE_IN_BYTES


def encode_request(name, keys, replace_existing_values):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    FixSizedTypesCodec.encode_boolean(buf, _REQUEST_REPLACE_EXISTING_VALUES_OFFSET, replace_existing_values)
    StringCodec.encode(buf, name)
    ListMultiFrameCodec.encode(buf, keys, DataCodec.encode, True)
    return OutboundMessage(buf, False, True)



from hazelcast.protocol.builtin import FixSizedTypesCodec, CodecUtil
from hazelcast.serialization.bits import *
from hazelcast.protocol.client_message import END_FRAME_BUF, END_FINAL_FRAME_BUF, SIZE_OF_FRAME_LENGTH_AND_FLAGS, create_initial_buffer_custom
from hazelcast.config import BitmapIndexOptions
from hazelcast.protocol.builtin import StringCodec

_UNIQUE_KEY_TRANSFORMATION_ENCODE_OFFSET = 2 * SIZE_OF_FRAME_LENGTH_AND_FLAGS
_UNIQUE_KEY_TRANSFORMATION_DECODE_OFFSET = 0
_INITIAL_FRAME_SIZE = _UNIQUE_KEY_TRANSFORMATION_ENCODE_OFFSET + INT_SIZE_IN_BYTES - SIZE_OF_FRAME_LENGTH_AND_FLAGS


class BitmapIndexOptionsCodec:
    @staticmethod
    def encode(buf, bitmap_index_options, is_final=False):
        initial_frame_buf = create_initial_buffer_custom(_INITIAL_FRAME_SIZE)
        FixSizedTypesCodec.encode_int(initial_frame_buf, _UNIQUE_KEY_TRANSFORMATION_ENCODE_OFFSET, bitmap_index_options.unique_key_transformation)
        buf.extend(initial_frame_buf)
        StringCodec.encode(buf, bitmap_index_options.unique_key)
        if is_final:
            buf.extend(END_FINAL_FRAME_BUF)
        else:
            buf.extend(END_FRAME_BUF)

    @staticmethod
    def decode(msg):
        msg.next_frame()
        initial_frame = msg.next_frame()
        unique_key_transformation = FixSizedTypesCodec.decode_int(initial_frame.buf, _UNIQUE_KEY_TRANSFORMATION_DECODE_OFFSET)
        unique_key = StringCodec.decode(msg)
        CodecUtil.fast_forward_to_end_frame(msg)
        return BitmapIndexOptions(unique_key, unique_key_transformation)



import typing

from hazelcast.future import Future
from hazelcast.protocol.codec import (
    atomic_ref_compare_and_set_codec,
    atomic_ref_get_codec,
    atomic_ref_set_codec,
    atomic_ref_contains_codec,
    atomic_ref_apply_codec,
)
from hazelcast.proxy.cp import BaseCPProxy
from hazelcast.types import ElementType
from hazelcast.serialization.compact import SchemaNotReplicatedError
from hazelcast.util import check_not_none


class AtomicReference(BaseCPProxy["BlockingAtomicReference"], typing.Generic[ElementType]):
    """A distributed, highly available object reference with atomic operations.

    AtomicReference offers linearizability during crash failures and network
    partitions. It is CP with respect to the CAP principle. If a network
    partition occurs, it remains available on at most one side of the
    partition.

    The following are some considerations you need to know when you use
    AtomicReference:

    - AtomicReference works based on the byte-content and not on the
      object-reference. If you use the ``compare_and_set()`` method, do not
      change the original value because its serialized content will then be
      different.
    - All methods returning an object return a private copy. You can modify the
      private copy, but the rest of the world is shielded from your changes. If
      you want these changes to be visible to the rest of the world, you need
      to write the change back to the AtomicReference; but be careful about
      introducing a data-race.
    - The in-memory format of an AtomicReference is ``binary``. The receiving
      side does not need to have the class definition available unless it needs
      to be deserialized on the other side., e.g., because a method like
      `alter()` is executed. This deserialization is done for every call that
      needs to have the object instead of the binary content, so be careful
      with expensive object graphs that need to be deserialized.
    - If you have an object with many fields or an object graph, and you only
      need to calculate some information or need a subset of fields, you can
      use the `apply()` method. With the `apply()` method, the whole object
      does not need to be sent over the line; only the information that is
      relevant is sent.

    IAtomicReference does not offer exactly-once / effectively-once
    execution semantics. It goes with at-least-once execution semantics
    by default and can cause an API call to be committed multiple times
    in case of CP member failures. It can be tuned to offer at-most-once
    execution semantics. Please see `fail-on-indeterminate-operation-state`
    server-side setting.
    """

    def compare_and_set(
        self, expect: typing.Optional[ElementType], update: typing.Optional[ElementType]
    ) -> Future[bool]:
        """Atomically sets the value to the given updated value
        only if the current value is equal to the expected value.

        Args:
            expect: The expected value.
            update: The new value.

        Returns:
            ``True`` if successful, or ``False`` if the actual value was not
            equal to the expected value.
        """
        try:
            expected_data = self._to_data(expect)
            new_data = self._to_data(update)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.compare_and_set, expect, update)

        codec = atomic_ref_compare_and_set_codec
        request = codec.encode_request(self._group_id, self._object_name, expected_data, new_data)
        return self._invoke(request, codec.decode_response)

    def get(self) -> Future[typing.Optional[ElementType]]:
        """Gets the current value.

        Returns:
            The current value.
        """
        codec = atomic_ref_get_codec
        request = codec.encode_request(self._group_id, self._object_name)

        def handler(response):
            return self._to_object(codec.decode_response(response))

        return self._invoke(request, handler)

    def set(self, new_value: typing.Optional[ElementType]) -> Future[None]:
        """Atomically sets the given value.

        Args:
            new_value: The new value.
        """
        try:
            new_value_data = self._to_data(new_value)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.set, new_value)

        codec = atomic_ref_set_codec
        request = codec.encode_request(self._group_id, self._object_name, new_value_data, False)
        return self._invoke(request)

    def get_and_set(
        self, new_value: typing.Optional[ElementType]
    ) -> Future[typing.Optional[ElementType]]:
        """Gets the old value and sets the new value.

        Args:
            new_value: The new value.

        Returns:
            The old value.
        """
        try:
            new_value_data = self._to_data(new_value)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.get_and_set, new_value)

        codec = atomic_ref_set_codec
        request = codec.encode_request(self._group_id, self._object_name, new_value_data, True)

        def handler(response):
            return self._to_object(codec.decode_response(response))

        return self._invoke(request, handler)

    def is_none(self) -> Future[bool]:
        """Checks if the stored reference is ``None``.

        Returns:
            ``True`` if the stored reference is ``None``, ``False`` otherwise.
        """
        return self.contains(None)

    def clear(self) -> Future[None]:
        """Clears the current stored reference, so it becomes ``None``."""
        return self.set(None)

    def contains(self, value: typing.Optional[ElementType]) -> Future[bool]:
        """Checks if the reference contains the value.

        Args:
            value: The value to check (is allowed to be ``None``).

        Returns:
            ``True`` if the value is found, ``False`` otherwise.
        """
        try:
            value_data = self._to_data(value)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.contains, value)

        codec = atomic_ref_contains_codec
        request = codec.encode_request(self._group_id, self._object_name, value_data)
        return self._invoke(request, codec.decode_response)

    def alter(self, function: typing.Any) -> Future[None]:
        """Alters the currently stored reference by applying a function on it.

        Notes:
            ``function`` must be an instance of Hazelcast serializable type.
            It must have a counterpart registered in the server-side that
            implements the ``com.hazelcast.core.IFunction`` interface with
            the actual logic of the function to be applied.

        Args:
            function: The function that alters the currently stored reference.
        """
        check_not_none(function, "Function cannot be None")
        try:
            function_data = self._to_data(function)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.alter, function)

        codec = atomic_ref_apply_codec
        # 0 means don't return the value
        request = codec.encode_request(self._group_id, self._object_name, function_data, 0, True)
        return self._invoke(request)

    def alter_and_get(self, function: typing.Any) -> Future[typing.Optional[ElementType]]:
        """Alters the currently stored reference by applying a function on it
        and gets the result.

        Notes:
            ``function`` must be an instance of Hazelcast serializable type.
            It must have a counterpart registered in the server-side that
            implements the ``com.hazelcast.core.IFunction`` interface with
            the actual logic of the function to be applied.

        Args:
            function: The function that alters the currently stored reference.

        Returns:
            The new value, the result of the applied function.
        """
        check_not_none(function, "Function cannot be None")
        try:
            function_data = self._to_data(function)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.alter_and_get, function)

        codec = atomic_ref_apply_codec
        # 2 means return the new value
        request = codec.encode_request(self._group_id, self._object_name, function_data, 2, True)

        def handler(response):
            return self._to_object(codec.decode_response(response))

        return self._invoke(request, handler)

    def get_and_alter(self, function: typing.Any) -> Future[typing.Optional[ElementType]]:
        """Alters the currently stored reference by applying a function on it
        on and gets the old value.

        Notes:
            ``function`` must be an instance of Hazelcast serializable type.
            It must have a counterpart registered in the server-side that
            implements the ``com.hazelcast.core.IFunction`` interface with
            the actual logic of the function to be applied.

        Args:
            function: The function that alters the currently stored reference.

        Returns:
            The old value, the value before the function is applied.
        """
        check_not_none(function, "Function cannot be None")
        try:
            function_data = self._to_data(function)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.get_and_alter, function)

        codec = atomic_ref_apply_codec
        # 1 means return the old value
        request = codec.encode_request(self._group_id, self._object_name, function_data, 1, True)

        def handler(response):
            return self._to_object(codec.decode_response(response))

        return self._invoke(request, handler)

    def apply(self, function: typing.Any) -> Future[typing.Optional[ElementType]]:
        """Applies a function on the value, the actual stored value will not
        change.

        Notes:
            ``function`` must be an instance of Hazelcast serializable type.
            It must have a counterpart registered in the server-side that
            implements the ``com.hazelcast.core.IFunction`` interface with
            the actual logic of the function to be applied.

        Args:
            function: The function applied on the currently stored reference.

        Returns:
            The result of the function application.
        """
        check_not_none(function, "Function cannot be None")
        try:
            function_data = self._to_data(function)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.apply, function)

        codec = atomic_ref_apply_codec
        # 2 means return the new value
        request = codec.encode_request(self._group_id, self._object_name, function_data, 2, False)

        def handler(response):
            return self._to_object(codec.decode_response(response))

        return self._invoke(request, handler)

    def blocking(self) -> "BlockingAtomicReference[ElementType]":
        return BlockingAtomicReference(self)


class BlockingAtomicReference(AtomicReference[ElementType]):
    __slots__ = ("_wrapped",)

    def __init__(self, wrapped: AtomicReference[ElementType]):
        self._wrapped = wrapped

    def compare_and_set(  # type: ignore[override]
        self,
        expect: typing.Optional[ElementType],
        update: typing.Optional[ElementType],
    ) -> bool:
        return self._wrapped.compare_and_set(expect, update).result()

    def get(  # type: ignore[override]
        self,
    ) -> typing.Optional[ElementType]:
        return self._wrapped.get().result()

    def set(  # type: ignore[override]
        self,
        new_value: typing.Optional[ElementType],
    ) -> None:
        return self._wrapped.set(new_value).result()

    def get_and_set(  # type: ignore[override]
        self,
        new_value: typing.Optional[ElementType],
    ) -> typing.Optional[ElementType]:
        return self._wrapped.get_and_set(new_value).result()

    def is_none(  # type: ignore[override]
        self,
    ) -> bool:
        return self._wrapped.is_none().result()

    def clear(  # type: ignore[override]
        self,
    ) -> None:
        return self._wrapped.clear().result()

    def contains(  # type: ignore[override]
        self,
        value: typing.Optional[ElementType],
    ) -> bool:
        return self._wrapped.contains(value).result()

    def alter(  # type: ignore[override]
        self,
        function: typing.Any,
    ) -> None:
        return self._wrapped.alter(function).result()

    def alter_and_get(  # type: ignore[override]
        self,
        function: typing.Any,
    ) -> typing.Optional[ElementType]:
        return self._wrapped.alter_and_get(function).result()

    def get_and_alter(  # type: ignore[override]
        self,
        function: typing.Any,
    ) -> typing.Optional[ElementType]:
        return self._wrapped.get_and_alter(function).result()

    def apply(  # type: ignore[override]
        self,
        function: typing.Any,
    ) -> typing.Optional[ElementType]:
        return self._wrapped.apply(function).result()

    def destroy(  # type: ignore[override]
        self,
    ) -> None:
        return self._wrapped.destroy().result()

    def blocking(self) -> "BlockingAtomicReference[ElementType]":
        return self

    def __repr__(self) -> str:
        return self._wrapped.__repr__()



import logging
import sys
import threading
import typing

from hazelcast.cluster import ClusterService, _InternalClusterService
from hazelcast.compact import CompactSchemaService
from hazelcast.config import Config
from hazelcast.connection import ConnectionManager, DefaultAddressProvider
from hazelcast.core import DistributedObjectEvent, DistributedObjectInfo
from hazelcast.cp import CPSubsystem, ProxySessionManager
from hazelcast.discovery import HazelcastCloudAddressProvider
from hazelcast.errors import IllegalStateError, InvalidConfigurationError
from hazelcast.future import Future
from hazelcast.invocation import InvocationService, Invocation
from hazelcast.lifecycle import LifecycleService, LifecycleState, _InternalLifecycleService
from hazelcast.listener import ClusterViewListenerService, ListenerService
from hazelcast.near_cache import NearCacheManager
from hazelcast.partition import PartitionService, _InternalPartitionService
from hazelcast.protocol.codec import (
    client_add_distributed_object_listener_codec,
    client_get_distributed_objects_codec,
    client_remove_distributed_object_listener_codec,
)
from hazelcast.proxy import (
    EXECUTOR_SERVICE,
    FLAKE_ID_GENERATOR_SERVICE,
    LIST_SERVICE,
    MAP_SERVICE,
    MULTI_MAP_SERVICE,
    PN_COUNTER_SERVICE,
    QUEUE_SERVICE,
    RELIABLE_TOPIC_SERVICE,
    REPLICATED_MAP_SERVICE,
    RINGBUFFER_SERVICE,
    SET_SERVICE,
    TOPIC_SERVICE,
    Executor,
    FlakeIdGenerator,
    List,
    MultiMap,
    PNCounter,
    ProxyManager,
    Queue,
    ReliableTopic,
    ReplicatedMap,
    Ringbuffer,
    Set,
    Topic,
)
from hazelcast.proxy.base import Proxy
from hazelcast.proxy.map import Map
from hazelcast.reactor import AsyncoreReactor
from hazelcast.serialization import SerializationServiceV1
from hazelcast.sql import SqlService, _InternalSqlService
from hazelcast.statistics import Statistics
from hazelcast.transaction import TWO_PHASE, Transaction, TransactionManager
from hazelcast.types import KeyType, ValueType, ItemType, MessageType
from hazelcast.util import AtomicInteger, RoundRobinLB

__all__ = ("HazelcastClient",)

_logger = logging.getLogger(__name__)


class HazelcastClient:
    """Hazelcast client instance to access and manipulate distributed data
    structures on the Hazelcast clusters.
    """

    _CLIENT_ID = AtomicInteger()

    def __init__(self, config: Config = None, **kwargs):
        """The client can be configured either by:

        - providing a configuration object as the first parameter of the
          constructor

        .. code:: python

            from hazelcast import HazelcastClient
            from hazelcast.config import Config

            config = Config()
            config.cluster_name = "a-cluster"
            client = HazelcastClient(config)

        - passing configuration options as keyword arguments

        .. code:: python

            from hazelcast import HazelcastClient

            client = HazelcastClient(
                cluster_name="a-cluster",
            )


        See the :class:`hazelcast.config.Config` documentation for the possible
        configuration options.

        Args:
            config: Optional configuration object.
            **kwargs: Optional keyword arguments of the client configuration.
        """
        if config:
            if kwargs:
                raise InvalidConfigurationError(
                    "Ambiguous client configuration is found. Either provide "
                    "the config object as the only parameter, or do not "
                    "pass it and use keyword arguments to configure the "
                    "client."
                )
        else:
            config = Config.from_dict(kwargs)

        self._config = config
        self._context = _ClientContext()
        client_id = HazelcastClient._CLIENT_ID.get_and_increment()
        self._name = self._create_client_name(client_id)
        self._reactor = AsyncoreReactor()
        self._serialization_service = SerializationServiceV1(config)
        self._near_cache_manager = NearCacheManager(config, self._serialization_service)
        self._internal_lifecycle_service = _InternalLifecycleService(config)
        self._lifecycle_service = LifecycleService(self._internal_lifecycle_service)
        self._internal_cluster_service = _InternalClusterService(self, config)
        self._cluster_service = ClusterService(self._internal_cluster_service)
        self._invocation_service = InvocationService(self, config, self._reactor)
        self._compact_schema_service = CompactSchemaService(
            self._serialization_service.compact_stream_serializer,
            self._invocation_service,
            self._cluster_service,
            self._reactor,
            self._config,
        )
        self._address_provider = self._create_address_provider()
        self._internal_partition_service = _InternalPartitionService(self)
        self._partition_service = PartitionService(
            self._internal_partition_service,
            self._serialization_service,
            self._compact_schema_service.send_schema_and_retry,
        )
        self._connection_manager = ConnectionManager(
            self,
            config,
            self._reactor,
            self._address_provider,
            self._internal_lifecycle_service,
            self._internal_partition_service,
            self._internal_cluster_service,
            self._invocation_service,
            self._near_cache_manager,
            self._send_state_to_cluster,
        )
        self._load_balancer = self._init_load_balancer(config)
        self._listener_service = ListenerService(
            self,
            config,
            self._connection_manager,
            self._invocation_service,
            self._compact_schema_service,
        )
        self._proxy_manager = ProxyManager(self._context)
        self._cp_subsystem = CPSubsystem(self._context)
        self._proxy_session_manager = ProxySessionManager(self._context)
        self._transaction_manager = TransactionManager(self._context)
        self._lock_reference_id_generator = AtomicInteger(1)
        self._statistics = Statistics(
            self,
            config,
            self._reactor,
            self._connection_manager,
            self._invocation_service,
            self._near_cache_manager,
        )
        self._cluster_view_listener = ClusterViewListenerService(
            self,
            self._connection_manager,
            self._internal_partition_service,
            self._internal_cluster_service,
            self._invocation_service,
        )
        self._shutdown_lock = threading.RLock()
        self._invocation_service.init(
            self._internal_partition_service,
            self._connection_manager,
            self._listener_service,
            self._compact_schema_service,
        )
        self._internal_sql_service = _InternalSqlService(
            self._connection_manager,
            self._serialization_service,
            self._invocation_service,
            self._compact_schema_service.send_schema_and_retry,
        )
        self._sql_service = SqlService(self._internal_sql_service)
        self._init_context()
        self._start()

    def _init_context(self):
        self._context.init_context(
            self,
            self._config,
            self._invocation_service,
            self._internal_partition_service,
            self._internal_cluster_service,
            self._connection_manager,
            self._serialization_service,
            self._listener_service,
            self._proxy_manager,
            self._near_cache_manager,
            self._lock_reference_id_generator,
            self._name,
            self._proxy_session_manager,
            self._reactor,
            self._compact_schema_service,
        )

    def _start(self):
        self._reactor.start()
        try:
            self._internal_lifecycle_service.start()
            self._invocation_service.start()
            membership_listeners = self._config.membership_listeners
            self._internal_cluster_service.start(self._connection_manager, membership_listeners)
            self._cluster_view_listener.start()
            self._connection_manager.start(self._load_balancer)
            sync_start = not self._config.async_start
            if sync_start:
                self._internal_cluster_service.wait_initial_member_list_fetched()
            self._connection_manager.connect_to_all_cluster_members(sync_start)

            self._listener_service.start()
            self._invocation_service.add_backup_listener()
            self._load_balancer.init(self._cluster_service)
            self._statistics.start()
        except:
            self.shutdown()
            raise
        _logger.info("Client started")

    def get_executor(self, name: str) -> Executor:
        """Creates cluster-wide ExecutorService.

        Args:
            name: Name of the Executor proxy.

        Returns:
            Executor proxy for the given name.
        """
        return self._proxy_manager.get_or_create(EXECUTOR_SERVICE, name)

    def get_flake_id_generator(self, name: str) -> FlakeIdGenerator:
        """Creates or returns a cluster-wide FlakeIdGenerator.

        Args:
            name: Name of the FlakeIdGenerator proxy.

        Returns:
            FlakeIdGenerator proxy for the given name.
        """
        return self._proxy_manager.get_or_create(FLAKE_ID_GENERATOR_SERVICE, name)

    def get_queue(self, name: str) -> Queue[ItemType]:
        """Returns the distributed queue instance with the specified name.

        Args:
            name: Name of the distributed queue.

        Returns:
            Distributed queue instance with the specified name.
        """
        return self._proxy_manager.get_or_create(QUEUE_SERVICE, name)

    def get_list(self, name: str) -> List[ItemType]:
        """Returns the distributed list instance with the specified name.

        Args:
            name: Name of the distributed list.

        Returns:
            Distributed list instance with the specified name.
        """
        return self._proxy_manager.get_or_create(LIST_SERVICE, name)

    def get_map(self, name: str) -> Map[KeyType, ValueType]:
        """Returns the distributed map instance with the specified name.

        Args:
            name: Name of the distributed map.

        Returns:
            Distributed map instance with the specified name.
        """
        return self._proxy_manager.get_or_create(MAP_SERVICE, name)

    def get_multi_map(self, name: str) -> MultiMap[KeyType, ValueType]:
        """Returns the distributed MultiMap instance with the specified name.

        Args:
            name: Name of the distributed MultiMap.

        Returns:
            Distributed MultiMap instance with the specified name.
        """
        return self._proxy_manager.get_or_create(MULTI_MAP_SERVICE, name)

    def get_pn_counter(self, name: str) -> PNCounter:
        """Returns the PN Counter instance with the specified name.

        Args:
            name: Name of the PN Counter.

        Returns:
            Distributed PN Counter instance with the specified name.
        """
        return self._proxy_manager.get_or_create(PN_COUNTER_SERVICE, name)

    def get_reliable_topic(self, name: str) -> ReliableTopic[MessageType]:
        """Returns the ReliableTopic instance with the specified name.

        Args:
            name: Name of the ReliableTopic.

        Returns:
            Distributed ReliableTopic instance with the specified name.
        """
        return self._proxy_manager.get_or_create(RELIABLE_TOPIC_SERVICE, name)

    def get_replicated_map(self, name: str) -> ReplicatedMap[KeyType, ValueType]:
        """Returns the distributed ReplicatedMap instance with the specified
        name.

        Args:
            name: Name of the distributed ReplicatedMap.

        Returns:
            Distributed ReplicatedMap instance with the specified name.
        """
        return self._proxy_manager.get_or_create(REPLICATED_MAP_SERVICE, name)

    def get_ringbuffer(self, name: str) -> Ringbuffer[ItemType]:
        """Returns the distributed Ringbuffer instance with the specified name.

        Args:
            name: Name of the distributed Ringbuffer.

        Returns:
            Distributed RingBuffer instance with the specified name.
        """

        return self._proxy_manager.get_or_create(RINGBUFFER_SERVICE, name)

    def get_set(self, name: str) -> Set[ItemType]:
        """Returns the distributed Set instance with the specified name.

        Args:
            name: Name of the distributed Set.

        Returns:
            Distributed Set instance with the specified name.
        """
        return self._proxy_manager.get_or_create(SET_SERVICE, name)

    def get_topic(self, name: str) -> Topic[MessageType]:
        """Returns the Topic instance with the specified name.

        Args:
            name: Name of the Topic.

        Returns:
            The Topic.
        """
        return self._proxy_manager.get_or_create(TOPIC_SERVICE, name)

    def new_transaction(
        self, timeout: float = 120, durability: int = 1, type: int = TWO_PHASE
    ) -> Transaction:
        """Creates a new Transaction associated with the current thread
         using default or given options.

        Args:
            timeout: The timeout in seconds determines the maximum lifespan of
                a transaction. So if a transaction is configured with a
                timeout of 2 minutes, then it will automatically rollback if
                it hasn't committed yet.
            durability: The durability is the number of machines that can take
                over if a member fails during a transaction commit or rollback.
            type: The transaction type which can be ``TWO_PHASE``
                or ``ONE_PHASE``.

        Returns:
            New Transaction associated with the current thread.
        """
        return self._transaction_manager.new_transaction(timeout, durability, type)

    def add_distributed_object_listener(
        self, listener_func: typing.Callable[[DistributedObjectEvent], None]
    ) -> Future[str]:
        """Adds a listener which will be notified when a new distributed object
        is created or destroyed.

        Args:
            listener_func: Function to be called when a distributed object is
                created or destroyed.

        Returns:
            A registration id which is used as a key to remove the listener.
        """
        is_smart = self._config.smart_routing
        codec = client_add_distributed_object_listener_codec
        request = codec.encode_request(is_smart)

        def handle_distributed_object_event(name, service_name, event_type, source):
            event = DistributedObjectEvent(name, service_name, event_type, source)
            listener_func(event)

        def event_handler(client_message):
            return codec.handle(client_message, handle_distributed_object_event)

        return self._listener_service.register_listener(
            request,
            codec.decode_response,
            client_remove_distributed_object_listener_codec.encode_request,
            event_handler,
        )

    def remove_distributed_object_listener(self, registration_id: str) -> Future[bool]:
        """Removes the specified distributed object listener.

        Returns silently if there is no such listener added before.

        Args:
            registration_id: The id of registered listener.

        Returns:
            ``True`` if registration is removed, ``False`` otherwise.
        """
        return self._listener_service.deregister_listener(registration_id)

    def get_distributed_objects(self) -> Future[typing.List[Proxy]]:
        """Returns all distributed objects such as; queue, map, set, list,
        topic, lock, multimap.

        Also, as a side effect, it clears the local instances of the destroyed
        proxies.

        Returns:
            List of instances created by Hazelcast.
        """
        request = client_get_distributed_objects_codec.encode_request()
        invocation = Invocation(request, response_handler=lambda m: m)
        self._invocation_service.invoke(invocation)

        local_distributed_object_infos = {
            DistributedObjectInfo(dist_obj.service_name, dist_obj.name)
            for dist_obj in self._proxy_manager.get_distributed_objects()
        }

        response = client_get_distributed_objects_codec.decode_response(invocation.future.result())
        for dist_obj_info in response:
            local_distributed_object_infos.discard(dist_obj_info)
            self._proxy_manager.get_or_create(
                dist_obj_info.service_name, dist_obj_info.name, create_on_remote=False
            )

        for dist_obj_info in local_distributed_object_infos:
            self._proxy_manager.destroy_proxy(
                dist_obj_info.service_name, dist_obj_info.name, destroy_on_remote=False
            )

        return self._proxy_manager.get_distributed_objects()

    def shutdown(self) -> None:
        """Shuts down this HazelcastClient."""
        with self._shutdown_lock:
            if self._internal_lifecycle_service.running:
                self._internal_lifecycle_service.fire_lifecycle_event(LifecycleState.SHUTTING_DOWN)
                self._internal_lifecycle_service.shutdown()
                self._proxy_session_manager.shutdown().result()
                self._near_cache_manager.destroy_near_caches()
                self._connection_manager.shutdown()
                self._invocation_service.shutdown()
                self._statistics.shutdown()
                self._reactor.shutdown()
                self._internal_lifecycle_service.fire_lifecycle_event(LifecycleState.SHUTDOWN)

    @property
    def name(self) -> str:
        """Name of the client."""
        return self._name

    @property
    def lifecycle_service(self) -> LifecycleService:
        """Lifecycle service allows you to check if the client is running and
        add and remove lifecycle listeners.
        """
        return self._lifecycle_service

    @property
    def partition_service(self) -> PartitionService:
        """Partition service allows you to get partition count, introspect
        the partition owners, and partition ids of keys.
        """
        return self._partition_service

    @property
    def cluster_service(self) -> ClusterService:
        """ClusterService: Cluster service allows you to get the list of
        the cluster members and add and remove membership listeners.
        """
        return self._cluster_service

    @property
    def cp_subsystem(self) -> CPSubsystem:
        """CP Subsystem offers set of in-memory linearizable data structures."""
        return self._cp_subsystem

    @property
    def sql(self) -> SqlService:
        """Returns a service to execute distributed SQL queries."""
        return self._sql_service

    def _create_address_provider(self):
        config = self._config
        cluster_members = config.cluster_members
        address_list_provided = len(cluster_members) > 0
        cloud_discovery_token = config.cloud_discovery_token
        cloud_enabled = cloud_discovery_token is not None
        if address_list_provided and cloud_enabled:
            raise IllegalStateError(
                "Only one discovery method can be enabled at a time. "
                "Cluster members given explicitly: %s, Hazelcast Viridian enabled: %s"
                % (address_list_provided, cloud_enabled)
            )

        if cloud_enabled:
            connection_timeout = self._get_connection_timeout(config)
            return HazelcastCloudAddressProvider(cloud_discovery_token, connection_timeout)

        return DefaultAddressProvider(cluster_members)

    def _create_client_name(self, client_id):
        client_name = self._config.client_name
        if client_name:
            return client_name
        return "hz.client_%s" % client_id

    def _send_state_to_cluster(self) -> Future:
        return self._compact_schema_service.send_all_schemas()

    @staticmethod
    def _get_connection_timeout(config):
        timeout = config.connection_timeout
        return sys.maxsize if timeout == 0 else timeout

    @staticmethod
    def _init_load_balancer(config):
        load_balancer = config.load_balancer
        if not load_balancer:
            load_balancer = RoundRobinLB()
        return load_balancer


class _ClientContext:
    """
    Context holding all the required services, managers and the configuration
    for a Hazelcast client.
    """

    def __init__(self):
        self.client = None
        self.config = None
        self.invocation_service = None
        self.partition_service = None
        self.cluster_service = None
        self.connection_manager = None
        self.serialization_service = None
        self.listener_service = None
        self.proxy_manager = None
        self.near_cache_manager = None
        self.lock_reference_id_generator = None
        self.name = None
        self.proxy_session_manager = None
        self.reactor = None
        self.compact_schema_service = None

    def init_context(
        self,
        client,
        config,
        invocation_service,
        partition_service,
        cluster_service,
        connection_manager,
        serialization_service,
        listener_service,
        proxy_manager,
        near_cache_manager,
        lock_reference_id_generator,
        name,
        proxy_session_manager,
        reactor,
        compact_schema_service,
    ):
        self.client = client
        self.config = config
        self.invocation_service = invocation_service
        self.partition_service = partition_service
        self.cluster_service = cluster_service
        self.connection_manager = connection_manager
        self.serialization_service = serialization_service
        self.listener_service = listener_service
        self.proxy_manager = proxy_manager
        self.near_cache_manager = near_cache_manager
        self.lock_reference_id_generator = lock_reference_id_generator
        self.name = name
        self.proxy_session_manager = proxy_session_manager
        self.reactor = reactor
        self.compact_schema_service = compact_schema_service



from hazelcast.protocol.builtin import CodecUtil
from hazelcast.protocol.client_message import END_FRAME_BUF, END_FINAL_FRAME_BUF, BEGIN_FRAME_BUF
from hazelcast.protocol.builtin import StringCodec
from hazelcast.serialization.compact import Schema
from hazelcast.protocol.builtin import ListMultiFrameCodec
from hazelcast.protocol.codec.custom.field_descriptor_codec import FieldDescriptorCodec


class SchemaCodec:
    @staticmethod
    def encode(buf, schema, is_final=False):
        buf.extend(BEGIN_FRAME_BUF)
        StringCodec.encode(buf, schema.type_name)
        ListMultiFrameCodec.encode(buf, schema.fields, FieldDescriptorCodec.encode)
        if is_final:
            buf.extend(END_FINAL_FRAME_BUF)
        else:
            buf.extend(END_FRAME_BUF)

    @staticmethod
    def decode(msg):
        msg.next_frame()
        type_name = StringCodec.decode(msg)
        fields = ListMultiFrameCodec.decode(msg, FieldDescriptorCodec.decode)
        CodecUtil.fast_forward_to_end_frame(msg)
        return Schema(type_name, fields)



import typing

from hazelcast.errors import IllegalStateError
from hazelcast.future import Future
from hazelcast.protocol.codec import (
    queue_add_all_codec,
    queue_add_listener_codec,
    queue_clear_codec,
    queue_compare_and_remove_all_codec,
    queue_compare_and_retain_all_codec,
    queue_contains_all_codec,
    queue_contains_codec,
    queue_drain_to_max_size_codec,
    queue_is_empty_codec,
    queue_iterator_codec,
    queue_offer_codec,
    queue_peek_codec,
    queue_poll_codec,
    queue_put_codec,
    queue_remaining_capacity_codec,
    queue_remove_codec,
    queue_remove_listener_codec,
    queue_size_codec,
    queue_take_codec,
)
from hazelcast.proxy.base import PartitionSpecificProxy, ItemEvent, ItemEventType
from hazelcast.types import ItemType
from hazelcast.serialization.compact import SchemaNotReplicatedError
from hazelcast.util import check_not_none, to_millis, deserialize_list_in_place


class Queue(PartitionSpecificProxy["BlockingQueue"], typing.Generic[ItemType]):
    """Concurrent, blocking, distributed, observable queue.

    Queue is not a partitioned data-structure. All of the Queue content is
    stored in a single machine (and in the backup). Queue will not scale by
    adding more members in the cluster.
    """

    def add(self, item: ItemType) -> Future[bool]:
        """Adds the specified item to this queue if there is available space.

        Args:
            item: The specified item.

        Returns:
            ``True`` if element is successfully added, ``False`` otherwise.
        """

        def result_fnc(f):
            if f.result():
                return True
            raise IllegalStateError("Queue is full!")

        return self.offer(item).continue_with(result_fnc)

    def add_all(self, items: typing.Sequence[ItemType]) -> Future[bool]:
        """Adds the elements in the specified collection to this queue.

        Args:
            items: Collection which includes the items to be added.

        Returns:
            ``True`` if this queue is changed after call, ``False`` otherwise.
        """
        check_not_none(items, "Value can't be None")
        try:
            data_items = []
            for item in items:
                check_not_none(item, "Value can't be None")
                data_items.append(self._to_data(item))
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.add_all, items)

        request = queue_add_all_codec.encode_request(self.name, data_items)
        return self._invoke(request, queue_add_all_codec.decode_response)

    def add_listener(
        self,
        include_value: bool = False,
        item_added_func: typing.Callable[[ItemEvent[ItemType]], None] = None,
        item_removed_func: typing.Callable[[ItemEvent[ItemType]], None] = None,
    ) -> Future[str]:
        """Adds an item listener for this queue. Listener will be notified for
         all queue add/remove events.

        Args:
            include_value: Whether received events include the updated item or
                not.
            item_added_func: Function to be called when an item is added to
                this set.
            item_removed_func: Function to be called when an item is deleted
                from this set.

        Returns:
            A registration id which is used as a key to remove the listener.
        """
        codec = queue_add_listener_codec
        request = codec.encode_request(self.name, include_value, self._is_smart)

        def handle_event_item(item_data, uuid, event_type):
            item = self._to_object(item_data) if include_value else None
            member = self._context.cluster_service.get_member(uuid)

            item_event = ItemEvent(self.name, item, event_type, member)
            if event_type == ItemEventType.ADDED:
                if item_added_func:
                    item_added_func(item_event)
            else:
                if item_removed_func:
                    item_removed_func(item_event)

        return self._register_listener(
            request,
            lambda r: codec.decode_response(r),
            lambda reg_id: queue_remove_listener_codec.encode_request(self.name, reg_id),
            lambda m: codec.handle(m, handle_event_item),
        )

    def clear(self) -> Future[None]:
        """Clears this queue. Queue will be empty after this call."""
        request = queue_clear_codec.encode_request(self.name)
        return self._invoke(request)

    def contains(self, item: ItemType) -> Future[bool]:
        """Determines whether this queue contains the specified item or not.

        Args:
            item: The specified item to be searched.

        Returns:
            ``True`` if the specified item exists in this queue, ``False``
            otherwise.
        """
        check_not_none(item, "Item can't be None")
        try:
            item_data = self._to_data(item)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.contains, item)

        request = queue_contains_codec.encode_request(self.name, item_data)
        return self._invoke(request, queue_contains_codec.decode_response)

    def contains_all(self, items: typing.Sequence[ItemType]) -> Future[bool]:
        """Determines whether this queue contains all of the items in the
        specified collection or not.

        Args:
            items: The specified collection which includes the items to be
                searched.

        Returns:
            ``True`` if all of the items in the specified collection exist in
            this queue, ``False`` otherwise.
        """
        check_not_none(items, "Items can't be None")
        try:
            data_items = []
            for item in items:
                check_not_none(item, "item can't be None")
                data_items.append(self._to_data(item))
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.contains_all, items)

        request = queue_contains_all_codec.encode_request(self.name, data_items)
        return self._invoke(request, queue_contains_all_codec.decode_response)

    def drain_to(self, target_list: typing.List[ItemType], max_size: int = -1) -> Future[int]:
        """Transfers all available items to the given `target_list` and removes
        these items from this queue.

        If a max_size is specified, it transfers at most the given number of
        items. In case of a failure, an item can exist in both collections or
        none of them.

        This operation may be more efficient than polling elements repeatedly
        and putting into collection.

        Args:
            target_list: the list where the items in this queue will be
                transferred.
            max_size: The maximum number items to transfer.

        Returns:
            Number of transferred items.
        """

        def handler(message):
            response = queue_drain_to_max_size_codec.decode_response(message)
            items = [self._to_object(item) for item in response]
            target_list.extend(items)
            return len(response)

        request = queue_drain_to_max_size_codec.encode_request(self.name, max_size)
        return self._invoke(request, handler)

    def iterator(self) -> Future[typing.List[ItemType]]:
        """Returns all the items in this queue.

        Returns:
            Collection of items in this queue.
        """

        def handler(message):
            data_list = queue_iterator_codec.decode_response(message)
            return deserialize_list_in_place(data_list, self._to_object)

        request = queue_iterator_codec.encode_request(self.name)
        return self._invoke(request, handler)

    def is_empty(self) -> Future[bool]:
        """Determines whether this set is empty or not.

        Returns:
            ``True`` if this queue is empty, ``False`` otherwise.
        """
        request = queue_is_empty_codec.encode_request(self.name)
        return self._invoke(request, queue_is_empty_codec.decode_response)

    def offer(self, item: ItemType, timeout: float = 0) -> Future[bool]:
        """Inserts the specified element into this queue if it is possible to
        do so immediately without violating capacity restrictions.

        If there is no space currently available:

        - If the timeout is provided, it waits until this timeout elapses
          and returns the result.
        - If the timeout is not provided, returns ``False`` immediately.

        Args:
            item: The item to be added.
            timeout: Maximum time in seconds to wait for addition.

        Returns:
            ``True`` if the element was added to this queue, ``False``
            otherwise.
        """
        check_not_none(item, "Value can't be None")
        try:
            element_data = self._to_data(item)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.offer, item, timeout)

        request = queue_offer_codec.encode_request(self.name, element_data, to_millis(timeout))
        return self._invoke(request, queue_offer_codec.decode_response)

    def peek(self) -> Future[typing.Optional[ItemType]]:
        """Retrieves the head of queue without removing it from the queue.

        Returns:
            The head of this queue, or ``None`` if this queue is empty.
        """

        def handler(message):
            return self._to_object(queue_peek_codec.decode_response(message))

        request = queue_peek_codec.encode_request(self.name)
        return self._invoke(request, handler)

    def poll(self, timeout: float = 0) -> Future[typing.Optional[ItemType]]:
        """Retrieves and removes the head of this queue.

        If this queue is empty:

        - If the timeout is provided, it waits until this timeout elapses
          and returns the result.
        - If the timeout is not provided, returns ``None``.

        Args:
            timeout: Maximum time in seconds to wait for addition.

        Returns:
            The head of this queue, or ``None`` if this queue is empty or
            specified timeout elapses before an item is added to the queue.
        """

        def handler(message):
            return self._to_object(queue_poll_codec.decode_response(message))

        request = queue_poll_codec.encode_request(self.name, to_millis(timeout))
        return self._invoke(request, handler)

    def put(self, item: ItemType) -> Future[None]:
        """Adds the specified element into this queue.

        If there is no space, it waits until necessary space becomes available.

        Args:
            item: The specified item.
        """
        check_not_none(item, "Value can't be None")
        try:
            element_data = self._to_data(item)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.put, item)

        request = queue_put_codec.encode_request(self.name, element_data)
        return self._invoke(request)

    def remaining_capacity(self) -> Future[int]:
        """Returns the remaining capacity of this queue.

        Returns:
            Remaining capacity of this queue.
        """
        request = queue_remaining_capacity_codec.encode_request(self.name)
        return self._invoke(request, queue_remaining_capacity_codec.decode_response)

    def remove(self, item: ItemType) -> Future[bool]:
        """Removes the specified element from the queue if it exists.

        Args:
            item: The specified element to be removed.

        Returns:
            ``True`` if the specified element exists in this queue, ``False``
            otherwise.
        """
        check_not_none(item, "Value can't be None")
        try:
            item_data = self._to_data(item)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.remove, item)

        request = queue_remove_codec.encode_request(self.name, item_data)
        return self._invoke(request, queue_remove_codec.decode_response)

    def remove_all(self, items: typing.Sequence[ItemType]) -> Future[bool]:
        """Removes all of the elements of the specified collection from this
        queue.

        Args:
            items: The specified collection.

        Returns:
            ``True`` if the call changed this queue, ``False`` otherwise.
        """
        check_not_none(items, "Value can't be None")
        try:
            data_items = []
            for item in items:
                check_not_none(item, "Value can't be None")
                data_items.append(self._to_data(item))
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.remove_all, items)

        request = queue_compare_and_remove_all_codec.encode_request(self.name, data_items)
        return self._invoke(request, queue_compare_and_remove_all_codec.decode_response)

    def remove_listener(self, registration_id: str) -> Future[bool]:
        """Removes the specified item listener.

        Returns silently if the specified listener was not added before.

        Args:
            registration_id: Id of the listener to be deleted.

        Returns:
            ``True`` if the item listener is removed, ``False`` otherwise.
        """
        return self._deregister_listener(registration_id)

    def retain_all(self, items: typing.Sequence[ItemType]) -> Future[bool]:
        """Removes the items which are not contained in the specified
        collection.

        In other words, only the items that are contained in the specified
        collection will be retained.

        Args:
            items: Collection which includes the elements to be retained in
                this set.

        Returns:
            ``True`` if this queue changed as a result of the call, ``False``
            otherwise.
        """
        check_not_none(items, "Value can't be None")
        try:
            data_items = []
            for item in items:
                check_not_none(item, "Value can't be None")
                data_items.append(self._to_data(item))
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.retain_all, items)

        request = queue_compare_and_retain_all_codec.encode_request(self.name, data_items)
        return self._invoke(request, queue_compare_and_retain_all_codec.decode_response)

    def size(self) -> Future[int]:
        """Returns the number of elements in this collection.

        If the size is greater than ``2**31 - 1``, it returns ``2**31 - 1``.

        Returns:
            Size of the queue.
        """
        request = queue_size_codec.encode_request(self.name)
        return self._invoke(request, queue_size_codec.decode_response)

    def take(self) -> Future[ItemType]:
        """Retrieves and removes the head of this queue, if necessary, waits
        until an item becomes available.

        Returns:
            The head of this queue.
        """

        def handler(message):
            return self._to_object(queue_take_codec.decode_response(message))

        request = queue_take_codec.encode_request(self.name)
        return self._invoke(request, handler)

    def blocking(self) -> "BlockingQueue[ItemType]":
        return BlockingQueue(self)


class BlockingQueue(Queue[ItemType]):
    __slots__ = ("_wrapped", "name", "service_name")

    def __init__(self, wrapped: Queue[ItemType]):
        self.name = wrapped.name
        self.service_name = wrapped.service_name
        self._wrapped = wrapped

    def add(  # type: ignore[override]
        self,
        item: ItemType,
    ) -> bool:
        return self._wrapped.add(item).result()

    def add_all(  # type: ignore[override]
        self,
        items: typing.Sequence[ItemType],
    ) -> bool:
        return self._wrapped.add_all(items).result()

    def add_listener(  # type: ignore[override]
        self,
        include_value: bool = False,
        item_added_func: typing.Callable[[ItemEvent[ItemType]], None] = None,
        item_removed_func: typing.Callable[[ItemEvent[ItemType]], None] = None,
    ) -> str:
        return self._wrapped.add_listener(
            include_value, item_added_func, item_removed_func
        ).result()

    def clear(  # type: ignore[override]
        self,
    ) -> None:
        return self._wrapped.clear().result()

    def contains(  # type: ignore[override]
        self,
        item: ItemType,
    ) -> bool:
        return self._wrapped.contains(item).result()

    def contains_all(  # type: ignore[override]
        self,
        items: typing.Sequence[ItemType],
    ) -> bool:
        return self._wrapped.contains_all(items).result()

    def drain_to(  # type: ignore[override]
        self,
        target_list: typing.List[ItemType],
        max_size: int = -1,
    ) -> int:
        return self._wrapped.drain_to(target_list, max_size).result()

    def iterator(  # type: ignore[override]
        self,
    ) -> typing.List[ItemType]:
        return self._wrapped.iterator().result()

    def is_empty(  # type: ignore[override]
        self,
    ) -> bool:
        return self._wrapped.is_empty().result()

    def offer(  # type: ignore[override]
        self,
        item: ItemType,
        timeout: float = 0,
    ) -> bool:
        return self._wrapped.offer(item, timeout).result()

    def peek(  # type: ignore[override]
        self,
    ) -> typing.Optional[ItemType]:
        return self._wrapped.peek().result()

    def poll(  # type: ignore[override]
        self,
        timeout: float = 0,
    ) -> typing.Optional[ItemType]:
        return self._wrapped.poll(timeout).result()

    def put(  # type: ignore[override]
        self,
        item: ItemType,
    ) -> None:
        return self._wrapped.put(item).result()

    def remaining_capacity(  # type: ignore[override]
        self,
    ) -> int:
        return self._wrapped.remaining_capacity().result()

    def remove(  # type: ignore[override]
        self,
        item: ItemType,
    ) -> bool:
        return self._wrapped.remove(item).result()

    def remove_all(  # type: ignore[override]
        self,
        items: typing.Sequence[ItemType],
    ) -> bool:
        return self._wrapped.remove_all(items).result()

    def remove_listener(  # type: ignore[override]
        self,
        registration_id: str,
    ) -> bool:
        return self._wrapped.remove_listener(registration_id).result()

    def retain_all(  # type: ignore[override]
        self,
        items: typing.Sequence[ItemType],
    ) -> bool:
        return self._wrapped.retain_all(items).result()

    def size(  # type: ignore[override]
        self,
    ) -> int:
        return self._wrapped.size().result()

    def take(  # type: ignore[override]
        self,
    ) -> ItemType:
        return self._wrapped.take().result()

    def destroy(self) -> bool:
        return self._wrapped.destroy()

    def blocking(self) -> "BlockingQueue[ItemType]":
        return self

    def __repr__(self) -> str:
        return self._wrapped.__repr__()



import logging
import typing

from hazelcast.errors import HazelcastSerializationError, IllegalStateError
from hazelcast.future import Future, ImmediateFuture
from hazelcast.invocation import Invocation
from hazelcast.protocol.codec import (
    client_fetch_schema_codec,
    client_send_schema_codec,
    client_send_all_schemas_codec,
)

if typing.TYPE_CHECKING:
    from hazelcast.cluster import ClusterService
    from hazelcast.config import Config
    from hazelcast.invocation import InvocationService
    from hazelcast.protocol.client_message import OutboundMessage
    from hazelcast.reactor import AsyncoreReactor
    from hazelcast.serialization.compact import (
        CompactStreamSerializer,
        Schema,
        SchemaNotReplicatedError,
    )

_logger = logging.getLogger(__name__)


class CompactSchemaService:
    _SEND_SCHEMA_RETRY_COUNT = 100

    def __init__(
        self,
        compact_serializer: "CompactStreamSerializer",
        invocation_service: "InvocationService",
        cluster_service: "ClusterService",
        reactor: "AsyncoreReactor",
        config: "Config",
    ):
        self._compact_serializer = compact_serializer
        self._invocation_service = invocation_service
        self._cluster_service = cluster_service
        self._reactor = reactor
        self._invocation_retry_pause = config.invocation_retry_pause
        self._has_replicated_schemas = False

    def fetch_schema(self, schema_id: int) -> Future:
        _logger.debug(
            "Could not find schema with the id %s locally. It will be fetched from the cluster.",
            schema_id,
        )

        request = client_fetch_schema_codec.encode_request(schema_id)
        fetch_schema_invocation = Invocation(
            request,
            response_handler=client_fetch_schema_codec.decode_response,
        )
        self._invocation_service.invoke(fetch_schema_invocation)
        return fetch_schema_invocation.future

    def send_schema_and_retry(
        self,
        error: "SchemaNotReplicatedError",
        func: typing.Callable[..., Future],
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> Future:
        schema = error.schema
        clazz = error.clazz
        request = client_send_schema_codec.encode_request(schema)

        def callback():
            self._has_replicated_schemas = True
            self._compact_serializer.register_schema_to_type(schema, clazz)
            return func(*args, **kwargs)

        return self._replicate_schema(
            schema, request, CompactSchemaService._SEND_SCHEMA_RETRY_COUNT, callback
        )

    def _replicate_schema(
        self,
        schema: "Schema",
        request: "OutboundMessage",
        remaining_retries: int,
        callback: typing.Callable[..., Future],
    ) -> Future:
        def continuation(future: Future):
            replicated_members = future.result()
            members = self._cluster_service.get_members()
            for member in members:
                if member.uuid not in replicated_members:
                    break
            else:
                # Loop completed normally.
                # All members in our member list all known to have the schema
                return callback()

            # There is a member in our member list that the schema
            # is not known to be replicated yet. We should retry
            # sending it in a random member.
            if remaining_retries <= 1:
                # We tried to send it a couple of times, but the member list
                # in our local and the member list returned by the initiator
                # nodes did not match.
                raise IllegalStateError(
                    f"The schema {schema} cannot be replicated in the cluster, "
                    f"after {CompactSchemaService._SEND_SCHEMA_RETRY_COUNT} retries. "
                    f"It might be the case that the client is connected to the two "
                    f"halves of the cluster that is experiencing a split-brain, "
                    f"and continue putting the data associated with that schema "
                    f"might result in data loss. It might be possible to replicate "
                    f"the schema after some time, when the cluster is healed."
                )

            delayed_future: Future = Future()
            self._reactor.add_timer(
                self._invocation_retry_pause,
                lambda: delayed_future.set_result(None),
            )

            def retry(_):
                return self._replicate_schema(
                    schema, request.copy(), remaining_retries - 1, callback
                )

            return delayed_future.continue_with(retry)

        return self._send_schema_replication_request(request).continue_with(continuation)

    def _send_schema_replication_request(self, request: "OutboundMessage") -> Future:
        invocation = Invocation(request, response_handler=client_send_schema_codec.decode_response)
        self._invocation_service.invoke(invocation)
        return invocation.future

    def send_all_schemas(self) -> Future:
        schemas = self._compact_serializer.get_schemas()
        if not schemas:
            _logger.debug("There is no schema to send to the cluster.")
            return ImmediateFuture(None)

        _logger.debug("Sending the following schemas to the cluster: %s", schemas)

        request = client_send_all_schemas_codec.encode_request(schemas)
        invocation = Invocation(request, urgent=True)
        self._invocation_service.invoke(invocation)
        return invocation.future

    def register_fetched_schema(self, schema_id: int, schema: typing.Optional["Schema"]) -> None:
        if not schema:
            raise HazelcastSerializationError(
                f"The schema with the id {schema_id} can not be found in the cluster."
            )

        self._compact_serializer.register_schema_to_id(schema)

    def has_replicated_schemas(self):
        """
        Returns ``True`` is the client has replicated
        any Compact schemas to the cluster.
        """
        return self._has_replicated_schemas



from hazelcast.hash import murmur_hash3_x86_32
from hazelcast.serialization import BE_INT
from hazelcast.serialization.serialization_const import *

PARTITION_HASH_OFFSET = 0
TYPE_OFFSET = 4
DATA_OFFSET = 8

HEAP_DATA_OVERHEAD = DATA_OFFSET


class Data:
    __slots__ = ("buffer",)
    """Data is basic unit of serialization.

    It stores binary form of an object serialized by serialization service.
    """

    def __init__(self, buf):
        self.buffer = buf

    def get_type(self):
        """Returns serialization type of binary form.

        Returns:
            int: Serialization type of binary form.
        """
        if len(self.buffer) == 0:
            return CONSTANT_TYPE_NULL
        return BE_INT.unpack_from(self.buffer, TYPE_OFFSET)[0]

    def total_size(self):
        """Returns the total size of Data in bytes.

        Returns:
            int: Total size of Data in bytes.
        """
        return len(self.buffer)

    def data_size(self):
        """Returns size of internal binary data in bytes.

        Returns:
            int: Size of internal binary data in bytes.
        """
        return max(len(self.buffer) - HEAP_DATA_OVERHEAD, 0)

    def get_partition_hash(self):
        """Returns partition hash calculated for serialized object.

        Partition hash is used to determine partition of a Data and is calculated using:

        - PartitioningStrategy during serialization.
        - If partition hash is not set then hash_code() is used.

        Returns:
            int: Partition hash.
        """
        partition_hash = BE_INT.unpack_from(self.buffer, PARTITION_HASH_OFFSET)[0]
        if partition_hash != 0:
            return partition_hash
        return self.hash_code()

    def is_portable(self):
        """Determines whether this Data is created from a ``Portable`` object or not.

        Returns:
            bool: ``True`` if source object is Portable, ``False`` otherwise.
        """
        return CONSTANT_TYPE_PORTABLE == self.get_type()

    def hash_code(self):
        """Returns the murmur hash of the internal data.

        Returns:
            int: The murmur hash of the internal data.
        """
        return murmur_hash3_x86_32(self.buffer)

    def __hash__(self):
        return murmur_hash3_x86_32(self.buffer)

    def __eq__(self, other):
        return (
            isinstance(other, Data)
            and self.total_size() == other.total_size()
            and self.buffer == other.buffer
        )

    def __len__(self):
        return self.total_size()



from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer
from hazelcast.protocol.codec.custom.schema_codec import SchemaCodec
from hazelcast.protocol.builtin import SetUUIDCodec

# hex: 0x001300
_REQUEST_MESSAGE_TYPE = 4864
# hex: 0x001301
_RESPONSE_MESSAGE_TYPE = 4865

_REQUEST_INITIAL_FRAME_SIZE = REQUEST_HEADER_SIZE


def encode_request(schema):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    SchemaCodec.encode(buf, schema, True)
    return OutboundMessage(buf, True)


def decode_response(msg):
    msg.next_frame()
    return SetUUIDCodec.decode(msg)



from datetime import date, datetime, time
from time import localtime
from typing import Any, Callable, Iterator, List, Optional, Sequence, Union, Tuple, Set, NamedTuple
import enum
import itertools
import threading
import urllib.parse

from hazelcast import HazelcastClient
from hazelcast.config import Config
from hazelcast.sql import (
    HazelcastSqlError,
    SqlColumnType,
    SqlResult,
    SqlRow,
    SqlRowMetadata,
    SqlExpectedResultType,
    _DEFAULT_CURSOR_BUFFER_SIZE,
)

apilevel = "2.0"
# Threads may share the module and connections.
threadsafety = 2
paramstyle = "qmark"


class Type(enum.Enum):
    """Type is the column type"""

    NULL = 0
    STRING = 1
    BOOLEAN = 2
    DATE = 3
    TIME = 4
    DATETIME = 5
    INTEGER = 6
    FLOAT = 7
    DECIMAL = 8
    JSON = 9
    OBJECT = 10


ColumnDescription = NamedTuple(
    "ColumnDescription",
    [
        ("name", str),
        ("type", Type),
        ("display_size", None),
        ("internal_size", None),
        ("precision", None),
        ("scale", None),
        ("null_ok", bool),
    ],
)
ColumnDescription.__doc__ = "ColumnDescription provides name, type and nullability information"


class _DBAPIType:
    def __init__(self, *values: Type):
        self._values = values

    def __eq__(self, other: object) -> bool:
        return other in self._values

    def __ne__(self, other: object) -> bool:
        return other not in self._values


Date = date
Time = time
Timestamp = datetime
Binary = bytes
STRING = _DBAPIType(Type.STRING)
DATETIME = _DBAPIType(Type.DATE, Type.TIME, Type.DATETIME)
BINARY = _DBAPIType()
NUMBER = _DBAPIType(Type.INTEGER, Type.FLOAT, Type.DECIMAL)
ROWID = _DBAPIType()


def DateFromTicks(ticks):
    return date(*localtime(ticks)[:3])


def TimeFromTicks(ticks):
    return time(*localtime(ticks)[3:6])


def TimestampFromTicks(ticks):
    return datetime(*localtime(ticks)[:6])


class Cursor:
    """Cursor is a database cursor object

    This class should not be initiated directly.
    Use connection.cursor() method to create one.
    """

    def __init__(self, conn: "Connection"):
        self.arraysize = 1
        self._conn = conn
        self._res: Union[SqlResult, None] = None
        self._description: Union[List[ColumnDescription], None] = None
        self._iter: Optional[Iterator[SqlRow]] = None
        self._rownumber = -1
        self._closed = False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __iter__(self) -> Optional[Iterator[SqlRow]]:
        return self._iter

    @property
    def connection(self) -> "Connection":
        """Returns the Connection object that created this cursor

        Returns:
            The Connection of this cursor
        """
        return self._conn

    @property
    def description(self) -> Union[List[ColumnDescription], None]:
        """Returns the descriptions of the columns

        Get the descriptions after calling execute.

        Returns:
            The list of column descriptions.
        """
        return self._description

    @property
    def rowcount(self) -> int:
        """Returns the number of rows in the result.

        This is not supported by this driver and always ``-1`` is returned.

        Returns:
            ``-1``
        """
        return -1

    @property
    def rownumber(self) -> Optional[int]:
        """Returns the index of the cursor in the result set

        Returns:
            0-based index of the cursor in the result set.
        """
        if self._rownumber < 0:
            return None
        return self._rownumber

    def close(self):
        """Closes the cursor and releases the resources"""
        if not self._closed:
            self._closed = True
            self._conn._close_cursor(self)
            if self._res:
                self._res.close()
                self._res = None

    def execute(self, operation: str, params: Optional[Tuple] = None) -> None:
        """Executes the given query with optional parameters

        Args:
            operation: A SQL string. Use question mark (``?``) as the
                placeholder if necessary.
            params: Optional tuple that contains the actual parameters
                to replace the placeholders in the query.
        """
        if params is not None and not isinstance(params, tuple):
            raise InterfaceError("params must be a tuple or None")
        params = params or ()
        self._ensure_open()
        self._rownumber = -1
        self._iter = None
        self._res = None
        cbs = _DEFAULT_CURSOR_BUFFER_SIZE
        if self.arraysize > 0:
            cbs = self.arraysize
        self._description = None
        res = _wrap_error(lambda: self._execute(operation, cbs, params))
        if res.is_row_set():
            self._rownumber = 0
            self._res = res
            self._description = self._make_description(res.get_row_metadata())
            self._iter = _IteratorWrapper(res)

    def _execute(self, operation, cbs, params):
        params = params or []
        return (
            self._conn._get_client()
            .sql.execute(operation, *params, cursor_buffer_size=cbs)
            .result()
        )

    def executemany(self, operation: str, seq_of_params: Sequence[Tuple]) -> None:
        """Runs the given query with the list of parameters

        Calling ``executemany(sql, [params1, params2, ...]`` is equivalent to
        ``execute(sql, params1), execute(sql, params2), ...``

        Args:
            operation: A SQL string. Use question mark (``?``) as the
                placeholder if necessary.
            seq_of_params: Optional list of tuples that contains the actual parameters
                to replace the placeholders in the query.
        """
        self._ensure_open()
        self._rownumber = -1
        self._iter = None
        self._res = None
        futures = []
        svc = self._conn._get_client().sql
        for params in seq_of_params:
            futures.append(
                svc.execute(
                    operation, *params, expected_result_type=SqlExpectedResultType.UPDATE_COUNT
                )
            )
        for fut in futures:
            _wrap_error(fut.result)

    def fetchone(self) -> Optional[SqlRow]:
        """Fetches a single row from the result

        Returns:
            A single row if there are rows in the result or ``None``.
        """
        if self._iter is None:
            raise InterfaceError("fetch can only be called after row returning queries")
        try:
            row = self.next()
            self._rownumber += 1
            return row
        except StopIteration:
            return None

    def fetchmany(self, size: Optional[int] = None) -> List[SqlRow]:
        """Fetches the given number of rows from the result

        Args:
            size: Optional number of rows to return.

        Returns:
            List of rows. The list will have at most `size` items.
        """
        if self._iter is None:
            raise InterfaceError("fetchmany can only be called after row returning queries")
        if size is None:
            size = self.arraysize
        rows = list(itertools.islice(self._iter, size))
        self._rownumber += len(rows)
        return rows

    def fetchall(self) -> List[SqlRow]:
        """Fetches all rows from the result

        This function should be called only with small and finite result sets.

        Returns:
            List of rows.
        """
        if self._iter is None:
            raise InterfaceError("fetchall can only be called after row returning queries")
        rows = list(self._iter)
        self._rownumber += len(rows)
        return rows

    def next(self) -> Optional[SqlRow]:
        if self._iter is None:
            return None
        return next(self._iter)

    def setinputsizes(self, sizes):
        pass

    def setoutputsize(self, size=None, column=None):
        pass

    @classmethod
    def _make_description(cls, metadata: SqlRowMetadata) -> List[ColumnDescription]:
        r = []
        for col in metadata.columns:
            r.append(
                ColumnDescription(
                    name=col.name,
                    type=_map_type(col.type),
                    display_size=None,
                    internal_size=None,
                    precision=None,
                    scale=None,
                    null_ok=col.nullable,
                )
            )
        return r

    def _ensure_open(self):
        if self._closed:
            raise self.connection.ProgrammingError("connection is closed")


class Connection:
    """Connection object provides connection to the Hazelcast cluster

    This class should not be initiated directly.
    Use ``connect`` method to create an instance.
    """

    def __init__(self, config: Config):
        self.__mu = threading.RLock()
        self.__client: Optional[HazelcastClient] = HazelcastClient(config)
        self._cursors: Set[Cursor] = set()

    def __enter__(self) -> "Connection":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close()

    def close(self) -> None:
        """Closes the connection and releases its resources"""

        if self.__client:
            with self.__mu:
                if self.__client:
                    self.__client.shutdown()
                    self.__client = None
                    return
        raise InterfaceError("connection was already closed")

    def commit(self) -> None:
        # transactions are not supported
        # ensure an exception is thrown if there is no client
        self._get_client()

    def cursor(self) -> Cursor:
        """Creates and returns a new cursor object

        Returns:
            Cursor object that uses this connection.
        """
        with self.__mu:
            if self.__client is not None:
                cursor = Cursor(self)
                self._cursors.add(cursor)
                return cursor
        raise InterfaceError("connection is already closed")

    def _get_client(self) -> HazelcastClient:
        with self.__mu:
            if self.__client is not None:
                return self.__client
        raise InterfaceError("connection is closed")

    def _close_cursor(self, cursor: Cursor) -> None:
        with self.__mu:
            if cursor in self._cursors:
                self._cursors.remove(cursor)

    @property
    def Error(self):
        return Error

    @property
    def Warning(self):
        return Warning

    @property
    def InterfaceError(self):
        return InterfaceError

    @property
    def DatabaseError(self):
        return DatabaseError

    @property
    def InternalError(self):
        return InternalError

    @property
    def OperationalError(self):
        return OperationalError

    @property
    def ProgrammingError(self):
        return ProgrammingError

    @property
    def IntegrityError(self):
        return IntegrityError

    @property
    def DataError(self):
        return DataError

    @property
    def NotSupportedError(self):
        return NotSupportedError


def connect(
    config=None,
    *,
    dsn="",
    user: str = None,
    password: str = None,
    host: str = None,
    port: int = None,
    cluster_name: str = None,
) -> Connection:
    """Creates a new Connection to the cluster

    Args:
         config: A Config object
         dsn: Dota Source Name in the following format:
            ``hz://[user:password]@addr1:port1[?opt1=value1[&opt2=value2 ...]]``
         user: Optional user name for authenticating to the cluster.
         password: Optional password for authenticating to the cluster.
         host: Hostname or IP address of the cluster. By default ``localhost``.
         port: Port of the cluster. By default ``5701``.
         cluster_name: Name of the cluster. By default ``dev``.

    Returns:
        Connection object.
    """
    c = _make_config(
        config,
        dsn=dsn,
        user=user,
        password=password,
        host=host,
        port=port,
        cluster_name=cluster_name,
    )
    return Connection(c)


class Error(Exception):
    pass


class Warning(Exception):
    pass


class InterfaceError(Error):
    pass


class DatabaseError(Error):
    pass


class InternalError(DatabaseError):
    pass


class OperationalError(DatabaseError):
    pass


class ProgrammingError(DatabaseError):
    pass


class IntegrityError(DatabaseError):
    pass


class DataError(DatabaseError):
    pass


class NotSupportedError(DatabaseError):
    pass


def _wrap_error(f: Callable) -> Any:
    try:
        return f()
    except HazelcastSqlError as e:
        raise DatabaseError(f"{e.args}") from e
    except StopIteration as e:
        raise e
    except Exception as e:
        raise DatabaseError from e


def _map_type(code: int) -> Type:
    type = _type_map.get(code)
    if type is None:
        raise NotSupportedError(f"Unknown type code: {code}")
    return type


_type_map = {
    SqlColumnType.VARCHAR: Type.STRING,
    SqlColumnType.BOOLEAN: Type.BOOLEAN,
    SqlColumnType.TINYINT: Type.INTEGER,
    SqlColumnType.SMALLINT: Type.INTEGER,
    SqlColumnType.INTEGER: Type.INTEGER,
    SqlColumnType.BIGINT: Type.INTEGER,
    SqlColumnType.DECIMAL: Type.DECIMAL,
    SqlColumnType.REAL: Type.FLOAT,
    SqlColumnType.DOUBLE: Type.FLOAT,
    SqlColumnType.DATE: Type.DATE,
    SqlColumnType.TIME: Type.TIME,
    SqlColumnType.TIMESTAMP: Type.DATETIME,
    SqlColumnType.TIMESTAMP_WITH_TIME_ZONE: Type.DATETIME,
    SqlColumnType.OBJECT: Type.OBJECT,
    SqlColumnType.NULL: Type.NULL,
    SqlColumnType.JSON: Type.JSON,
}


def _make_config(
    config: Config = None,
    *,
    dsn="",
    user: str = None,
    password: str = None,
    host: str = None,
    port: int = None,
    cluster_name: str = None,
) -> Config:
    kwargs_used = user or password or host or port or cluster_name
    if config is not None:
        if not isinstance(config, Config):
            raise InterfaceError("config must be a hazelcast.Config object")
        if dsn or kwargs_used:
            raise InterfaceError("config argument cannot be used with keyword arguments")
        return config
    if dsn:
        if kwargs_used:
            raise InterfaceError("dsn argument cannot be used with other keyword arguments")
        return _parse_dsn(dsn)
    config = Config()
    if not host:
        host = "localhost"
    if not port:
        port = 5701
    host = f"{host}:{port}"
    config.cluster_members = [host]
    if user is not None:
        config.creds_username = user
    if password is not None:
        config.creds_password = password
    if cluster_name is not None:
        config.cluster_name = cluster_name
    return config


def _parse_dsn(dsn: str) -> Config:
    r = urllib.parse.urlparse(dsn)
    if r.scheme != "hz":
        raise InterfaceError(f"Scheme must be hz, but it is: {r.scheme}")
    cfg = Config()
    host = "localhost"
    port = 5701
    if r.hostname:
        host = r.hostname
    if r.port:
        port = r.port
    cfg.cluster_members = [f"{host}:{port}"]
    if r.username:
        cfg.creds_username = r.username
    if r.password:
        cfg.creds_password = r.password
    for k, v in urllib.parse.parse_qsl(r.query):
        value: Any = v
        if k in _parse_dsn_map:
            attr_name, transform = _parse_dsn_map[k]
            if transform:
                try:
                    value = transform(value)
                except ValueError as e:
                    raise InterfaceError from e
            setattr(cfg, attr_name, value)
        else:
            raise InterfaceError(f"Unknown DSN attribute: {k}")
    return cfg


def _make_bool(v: str) -> bool:
    if v == "true":
        return True
    if v == "false":
        return False
    raise ValueError(f"Invalid boolean: {v}")


_parse_dsn_map = {
    "cluster.name": ("cluster_name", None),
    "cloud.token": ("cloud_discovery_token", None),
    "smart": ("smart_routing", _make_bool),
    "ssl": ("ssl_enabled", _make_bool),
    "ssl.ca.path": ("ssl_cafile", None),
    "ssl.cert.path": ("ssl_certfile", None),
    "ssl.key.path": ("ssl_keyfile", None),
    "ssl.key.password": ("ssl_password", None),
}


class _IteratorWrapper:
    def __init__(self, it: Iterator):
        self.it = it.__iter__()

    def __iter__(self):
        return self

    def __next__(self):
        return _wrap_error(lambda: next(self.it))



import typing
from uuid import uuid4
from hazelcast import future
from hazelcast.core import MemberInfo
from hazelcast.future import Future
from hazelcast.protocol.codec import (
    executor_service_shutdown_codec,
    executor_service_is_shutdown_codec,
    executor_service_submit_to_partition_codec,
    executor_service_submit_to_member_codec,
)
from hazelcast.proxy.base import Proxy
from hazelcast.serialization.compact import SchemaNotReplicatedError
from hazelcast.util import check_not_none


class Executor(Proxy["BlockingExecutor"]):
    """An object that executes submitted executable tasks."""

    def execute_on_key_owner(self, key: typing.Any, task: typing.Any) -> Future[typing.Any]:
        """Executes a task on the owner of the specified key.

        Args:
            key: The specified key.
            task: A task executed on the owner of the specified key.

        Returns:
            The result of the task.
        """
        check_not_none(key, "key can't be None")
        check_not_none(task, "task can't be None")

        try:
            key_data = self._to_data(key)
            task_data = self._to_data(task)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.execute_on_key_owner, key, task)

        partition_id = self._context.partition_service.get_partition_id(key_data)
        uuid = uuid4()

        def handler(message):
            return self._to_object(
                executor_service_submit_to_partition_codec.decode_response(message)
            )

        request = executor_service_submit_to_partition_codec.encode_request(
            self.name, uuid, task_data
        )
        return self._invoke_on_partition(request, partition_id, handler)

    def execute_on_member(self, member: MemberInfo, task: typing.Any) -> Future[typing.Any]:
        """Executes a task on the specified member.

        Args:
            member: The specified member.
            task: The task executed on the specified member.

        Returns:
            The result of the task.
        """
        check_not_none(task, "task can't be None")
        try:
            task_data = self._to_data(task)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.execute_on_member, member, task)

        uuid = uuid4()
        return self._execute_on_member(uuid, task_data, member.uuid)

    def execute_on_members(
        self, members: typing.Sequence[MemberInfo], task: typing.Any
    ) -> Future[typing.List[typing.Any]]:
        """Executes a task on each of the specified members.

        Args:
            members: The specified members.
            task: The task executed on the specified members.

        Returns:
            The list of results of the tasks on each member.
        """
        try:
            task_data = self._to_data(task)
        except SchemaNotReplicatedError as e:
            return self._send_schema_and_retry(e, self.execute_on_members, members, task)

        futures = []
        uuid = uuid4()
        for member in members:
            f = self._execute_on_member(uuid, task_data, member.uuid)
            futures.append(f)

        return future.combine_futures(futures)

    def execute_on_all_members(self, task: typing.Any) -> Future[typing.List[typing.Any]]:
        """Executes a task on all the known cluster members.

        Args:
            task: The task executed on the all the members.

        Returns:
            The list of results of the tasks on each member.
        """
        return self.execute_on_members(self._context.cluster_service.get_members(), task)

    def is_shutdown(self) -> Future[bool]:
        """Determines whether this executor has been shutdown or not.

        Returns:
            ``True`` if the executor has been shutdown, ``False`` otherwise.
        """
        request = executor_service_is_shutdown_codec.encode_request(self.name)
        return self._invoke(request, executor_service_is_shutdown_codec.decode_response)

    def shutdown(self) -> Future[None]:
        """Initiates a shutdown process which works orderly. Tasks that were
        submitted before shutdown are executed but new task will not be
        accepted.
        """
        request = executor_service_shutdown_codec.encode_request(self.name)
        return self._invoke(request)

    def _execute_on_member(self, uuid, task_data, member_uuid):
        def handler(message):
            return self._to_object(executor_service_submit_to_member_codec.decode_response(message))

        request = executor_service_submit_to_member_codec.encode_request(
            self.name, uuid, task_data, member_uuid
        )
        return self._invoke_on_target(request, member_uuid, handler)

    def blocking(self) -> "BlockingExecutor":
        return BlockingExecutor(self)


class BlockingExecutor(Executor):
    __slots__ = ("_wrapped", "name", "service_name")

    def __init__(self, wrapped: Executor):
        self.name = wrapped.name
        self.service_name = wrapped.service_name
        self._wrapped = wrapped

    def execute_on_key_owner(  # type: ignore[override]
        self,
        key: typing.Any,
        task: typing.Any,
    ) -> typing.Any:
        return self._wrapped.execute_on_key_owner(key, task).result()

    def execute_on_member(  # type: ignore[override]
        self,
        member: MemberInfo,
        task: typing.Any,
    ) -> typing.Any:
        return self._wrapped.execute_on_member(member, task).result()

    def execute_on_members(  # type: ignore[override]
        self,
        members: typing.Sequence[MemberInfo],
        task: typing.Any,
    ) -> typing.List[typing.Any]:
        return self._wrapped.execute_on_members(members, task).result()

    def execute_on_all_members(  # type: ignore[override]
        self,
        task: typing.Any,
    ) -> typing.List[typing.Any]:
        return self._wrapped.execute_on_all_members(task).result()

    def is_shutdown(  # type: ignore[override]
        self,
    ) -> bool:
        return self._wrapped.is_shutdown().result()

    def shutdown(  # type: ignore[override]
        self,
    ) -> None:
        return self._wrapped.shutdown().result()

    def blocking(self) -> "BlockingExecutor":
        return self

    def destroy(self) -> bool:
        return self._wrapped.destroy()

    def __repr__(self) -> str:
        return self._wrapped.__repr__()



import abc
import typing
import uuid

from hazelcast.core import MemberInfo
from hazelcast.types import KeyType, ValueType, ItemType, MessageType, BlockingProxyType
from hazelcast.invocation import Invocation
from hazelcast.partition import string_partition_strategy
from hazelcast.util import get_attr_name

MAX_SIZE = float("inf")


def _no_op_response_handler(_):
    return None


class Proxy(typing.Generic[BlockingProxyType], abc.ABC):
    """Provides basic functionality for Hazelcast Proxies."""

    def __init__(self, service_name: str, name: str, context):
        self.service_name = service_name
        self.name = name
        self._context = context
        self._invocation_service = context.invocation_service
        self._partition_service = context.partition_service
        serialization_service = context.serialization_service
        self._to_object = serialization_service.to_object
        self._to_data = serialization_service.to_data
        listener_service = context.listener_service
        self._register_listener = listener_service.register_listener
        self._deregister_listener = listener_service.deregister_listener
        self._is_smart = context.config.smart_routing
        self._send_schema_and_retry = context.compact_schema_service.send_schema_and_retry

    def destroy(self) -> bool:
        """Destroys this proxy.

        Returns:
            ``True`` if this proxy is destroyed successfully, ``False``
            otherwise.
        """
        self._on_destroy()
        return self._context.proxy_manager.destroy_proxy(self.service_name, self.name)

    def _on_destroy(self):
        pass

    def __repr__(self) -> str:
        return '%s(name="%s")' % (type(self).__name__, self.name)

    def _invoke(self, request, response_handler=_no_op_response_handler):
        invocation = Invocation(request, response_handler=response_handler)
        self._invocation_service.invoke(invocation)
        return invocation.future

    def _invoke_on_target(self, request, uuid, response_handler=_no_op_response_handler):
        invocation = Invocation(request, uuid=uuid, response_handler=response_handler)
        self._invocation_service.invoke(invocation)
        return invocation.future

    def _invoke_on_key(self, request, key_data, response_handler=_no_op_response_handler):
        partition_id = self._partition_service.get_partition_id(key_data)
        invocation = Invocation(
            request, partition_id=partition_id, response_handler=response_handler
        )
        self._invocation_service.invoke(invocation)
        return invocation.future

    def _invoke_on_partition(self, request, partition_id, response_handler=_no_op_response_handler):
        invocation = Invocation(
            request, partition_id=partition_id, response_handler=response_handler
        )
        self._invocation_service.invoke(invocation)
        return invocation.future

    @abc.abstractmethod
    def blocking(self) -> BlockingProxyType:
        """Returns a version of this proxy with only blocking method calls."""
        pass


class PartitionSpecificProxy(Proxy[BlockingProxyType], abc.ABC):
    """Provides basic functionality for Partition Specific Proxies."""

    def __init__(self, service_name, name, context):
        super(PartitionSpecificProxy, self).__init__(service_name, name, context)
        partition_key = context.serialization_service.to_data(string_partition_strategy(self.name))
        self._partition_id = context.partition_service.get_partition_id(partition_key)

    def _invoke(self, request, response_handler=_no_op_response_handler):
        invocation = Invocation(
            request, partition_id=self._partition_id, response_handler=response_handler
        )
        self._invocation_service.invoke(invocation)
        return invocation.future


class TransactionalProxy:
    """Provides an interface for all transactional distributed objects."""

    def __init__(self, name, transaction, context):
        self.name = name
        self.transaction = transaction
        self._invocation_service = context.invocation_service
        serialization_service = context.serialization_service
        self._to_object = serialization_service.to_object
        self._to_data = serialization_service.to_data
        self._send_schema_and_retry = context.compact_schema_service.send_schema_and_retry

    def _send_schema(self, error):
        return self._send_schema_and_retry(error, lambda: None).result()

    def _invoke(self, request, response_handler=_no_op_response_handler):
        invocation = Invocation(
            request, connection=self.transaction.connection, response_handler=response_handler
        )
        self._invocation_service.invoke(invocation)
        return invocation.future.result()

    def __repr__(self):
        return '%s(name="%s")' % (type(self).__name__, self.name)


class ItemEventType:
    """Type of item events."""

    ADDED = 1
    """
    Fired when an item is added.
    """

    REMOVED = 2
    """
    Fired when an item is removed.
    """


class EntryEventType:
    """Type of entry event."""

    ADDED = 1
    """
    Fired if an entry is added.
    """

    REMOVED = 2
    """
    Fired if an entry is removed.
    """

    UPDATED = 4
    """
    Fired if an entry is updated.
    """

    EVICTED = 8
    """
    Fired if an entry is evicted.
    """

    EXPIRED = 16
    """
    Fired if an entry is expired.
    """

    EVICT_ALL = 32
    """
    Fired if all entries are evicted.
    """

    CLEAR_ALL = 64
    """
    Fired if all entries are cleared.
    """

    MERGED = 128
    """
    Fired if an entry is merged after a network partition.
    """

    INVALIDATION = 256
    """
    Fired if an entry is invalidated.
    """

    LOADED = 512
    """
    Fired if an entry is loaded.
    """


class ItemEvent(typing.Generic[ItemType]):
    """Map Item event.

    Attributes:
        name: Name of the proxy that fired the event.
        item: The item related to the event.
        event_type: Type of the event.
        member: Member that fired the event.
    """

    def __init__(self, name: str, item: ItemEventType, event_type: int, member: MemberInfo):
        self.name = name
        self.item = item
        self.event_type = event_type
        self.member = member


class EntryEvent(typing.Generic[KeyType, ValueType]):
    """Map Entry event.

    Attributes:
        event_type: Type of the event.
        uuid: UUID of the member that fired the event.
        number_of_affected_entries: Number of affected entries by this event.
        key: The key of this entry event.
        value: The value of the entry event.
        old_value: The old value of the entry event.
        merging_value: The incoming merging value of the entry event.
    """

    def __init__(
        self,
        key: KeyType,
        value: ValueType,
        old_value: ValueType,
        merging_value: ValueType,
        event_type: int,
        member_uuid: uuid.UUID,
        number_of_affected_entries: int,
    ):
        self.key = key
        self.value = value
        self.old_value = old_value
        self.merging_value = merging_value
        self.event_type = event_type
        self.uuid = member_uuid
        self.number_of_affected_entries = number_of_affected_entries

    def __repr__(self):
        return (
            "EntryEvent(key=%s, value=%s, old_value=%s, merging_value=%s, event_type=%s, uuid=%s, "
            "number_of_affected_entries=%s)"
            % (
                self.key,
                self.value,
                self.old_value,
                self.merging_value,
                get_attr_name(EntryEventType, self.event_type),
                self.uuid,
                self.number_of_affected_entries,
            )
        )


class TopicMessage(typing.Generic[MessageType]):
    """Topic message.

    Attributes:
        name: Name of the proxy that fired the event.
        message: The message sent to Topic.
        publish_time: UNIX time that the event is published as seconds.
        member: Member that fired the event.
    """

    __slots__ = ("name", "message", "publish_time", "member")

    def __init__(self, name: str, message: MessageType, publish_time: int, member: MemberInfo):
        self.name = name
        self.message = message
        self.publish_time = publish_time
        self.member = member

    def __repr__(self):
        return "TopicMessage(message=%s, publish_time=%s, topic_name=%s, publishing_member=%s)" % (
            self.message,
            self.publish_time,
            self.name,
            self.member,
        )


def get_entry_listener_flags(**kwargs):
    flags = 0
    for key, value in kwargs.items():
        if value:
            flags |= getattr(EntryEventType, key)
    return flags
